"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import time
from operator import itemgetter

from boto3 import client, resource
from boto3.dynamodb.conditions import Attr, Key
from botocore.exceptions import ClientError

from syndicate.exceptions import ResourceProcessingError, \
    ParameterError
from syndicate.commons.log_helper import get_logger
from syndicate.connection.helper import apply_methods_decorator, retry


_LOG = get_logger(__name__)

DEFAULT_READ_CAPACITY = DEFAULT_WRITE_CAPACITY = 5
DEFAULT_GI_READ_CAPACITY = DEFAULT_GI_WRITE_CAPACITY = 1

UNLIMITED_REQUEST_UNIT = -1


def _append_attr_definition(definition, attr_name, attr_type):
    """ Adds an attribute definition if it is not already present.
    Dynamodb table creation requires to define attributes that are keys
    for the table or indexes. There must be no duplicated attribute
    definitions - aws throws ValidationException in such case.

    :type definition: []
    :type attr_name: str
    :type attr_type: str
    """
    for each in definition:
        if each['AttributeName'] == attr_name:
            return  # attribute is already defined, must not duplicate
    definition.append(dict(AttributeName=attr_name,
                           AttributeType=attr_type))


def _build_global_index_definition(
        index, billing_mode, read_throughput=None, write_throughput=None):
    index_info = _build_index_definition(index)

    if billing_mode == 'PAY_PER_REQUEST':
        index_info['OnDemandThroughput'] = dict(
            MaxReadRequestUnits=read_throughput or UNLIMITED_REQUEST_UNIT,
            MaxWriteRequestUnits=write_throughput or UNLIMITED_REQUEST_UNIT
        )
    else:
        index_info['ProvisionedThroughput'] = dict(
            ReadCapacityUnits=read_throughput or DEFAULT_GI_READ_CAPACITY,
            WriteCapacityUnits=write_throughput or DEFAULT_GI_WRITE_CAPACITY
        )

    return index_info


def _add_index_keys_to_definition(definition, index):
    _append_attr_definition(definition, index["index_key_name"],
                            index["index_key_type"])
    if index.get('index_sort_key_name'):
        _append_attr_definition(definition,
                                index["index_sort_key_name"],
                                index["index_sort_key_type"])


def _build_index_definition(index):
    """
    Creates request object to Index to be deployed
    :param index:
    :return:
    """
    index_def = {
        "IndexName": index["name"],
        "KeySchema": [
            {
                "AttributeName": index["index_key_name"],
                "KeyType": "HASH"
            }
        ],
        "Projection": {
            "ProjectionType": "ALL"
        }
    }
    if index.get('index_sort_key_name'):
        index_def['KeySchema'].append(
            {
                "AttributeName": index["index_sort_key_name"],
                "KeyType": "RANGE"
            })
    return index_def


@apply_methods_decorator(retry())
class DynamoConnection(object):
    """ DynamoDB class."""

    def __init__(self, region=None, aws_access_key_id=None,
                 aws_secret_access_key=None, endpoint=None,
                 aws_session_token=None):
        self.conn = resource("dynamodb", region,
                             aws_access_key_id=aws_access_key_id,
                             aws_secret_access_key=aws_secret_access_key,
                             endpoint_url=endpoint,
                             aws_session_token=aws_session_token)
        self.client = client("dynamodb", region,
                             aws_access_key_id=aws_access_key_id,
                             aws_secret_access_key=aws_secret_access_key,
                             endpoint_url=endpoint,
                             aws_session_token=aws_session_token)
        _LOG.debug('Opened new DynamoDB connection.')

    def create_table(self, table_name, hash_key_name, hash_key_type,
                     billing_mode, sort_key_name=None, sort_key_type=None,
                     read_throughput=None, write_throughput=None,
                     wait=True, global_indexes=None, local_indexes=None,
                     tags=None):
        """ Table creation.

        :type table_name: str
        :type hash_key_name: str
        :type hash_key_type: N/S/B
        :type sort_key_name: str
        :type sort_key_type: N/S/B
        :type billing_mode: PROVISIONED/PAY_PER_REQUEST
        :type read_throughput: int
        :type write_throughput: int
        :type wait: bool
        :type global_indexes: dict
        :type local_indexes: dict
        :type tags: list of dict
        :returns created table
        """

        params = dict()

        params['BillingMode'] = billing_mode
        if billing_mode == 'PAY_PER_REQUEST':
            params['OnDemandThroughput'] = dict(
                MaxReadRequestUnits=read_throughput or UNLIMITED_REQUEST_UNIT,
                MaxWriteRequestUnits=write_throughput or UNLIMITED_REQUEST_UNIT
            )
        else:
            params['ProvisionedThroughput'] = dict(
                ReadCapacityUnits=read_throughput or DEFAULT_READ_CAPACITY,
                WriteCapacityUnits=write_throughput or DEFAULT_WRITE_CAPACITY)

        schema = [dict(AttributeName=hash_key_name, KeyType='HASH')]
        definition = [dict(AttributeName=hash_key_name,
                           AttributeType=hash_key_type)]
        stream = {'StreamEnabled': False}
        if sort_key_name:
            schema.append(dict(AttributeName=sort_key_name,
                               KeyType='RANGE'))
            definition.append(dict(AttributeName=sort_key_name,
                                   AttributeType=sort_key_type))
        if global_indexes:
            for index in global_indexes:
                _add_index_keys_to_definition(definition=definition,
                                              index=index)
        if local_indexes:
            for index in local_indexes:
                _add_index_keys_to_definition(definition=definition,
                                              index=index)
        params.update(dict(
            TableName=table_name, KeySchema=schema,
            AttributeDefinitions=definition, StreamSpecification=stream
        ))
        if global_indexes:
            params['GlobalSecondaryIndexes'] = []
            for index in global_indexes:
                read_capacity = \
                    index.get('read_capacity') or read_throughput
                write_capacity = \
                    index.get('write_capacity') or write_throughput
                index_info = _build_global_index_definition(index,
                                                            billing_mode,
                                                            read_capacity,
                                                            write_capacity)
                params['GlobalSecondaryIndexes'].append(index_info)
        if local_indexes:
            params['LocalSecondaryIndexes'] = []
            for index in local_indexes:
                index_info = _build_index_definition(index)
                params['LocalSecondaryIndexes'].append(index_info)
        if tags:
            params['Tags'] = tags
        table = self.conn.create_table(**params)
        if wait:
            waiter = table.meta.client.get_waiter('table_exists')
            waiter.wait(TableName=table_name)
        return table

    def update_table_ttl(self, table_name, ttl_attribute_name, wait=True):
        """ Updates table ttl attribute

        :param table_name: DynamoDB table name
        :type table_name: str
        :param ttl_attribute_name: name of the table's attribute that holds
            ttl value
        :type ttl_attribute_name: str
        :param wait: to wait for table update to finish or not
        :type wait: bool
        :returns update_time_to_live response as dict
        """
        ttl_enabled = ttl_attribute_name is not None
        existing_ttl_specs = self.client.describe_time_to_live(
            TableName=table_name
        )['TimeToLiveDescription']
        existing_ttl_enabled = \
            existing_ttl_specs.get('TimeToLiveStatus') == 'ENABLED'
        existing_ttl_attribute_name = \
            existing_ttl_specs.get('AttributeName')

        if ttl_enabled != existing_ttl_enabled:
            _LOG.info(
                'Updating ttl value to {0} for table {1}, attribute name: {2}'
                .format(ttl_enabled, table_name, ttl_attribute_name))
            response = self.client.update_time_to_live(
                TableName=table_name,
                TimeToLiveSpecification={
                    'Enabled': ttl_enabled,
                    # set ttl attribute name to the one defined in meta when
                    # enabling ttl, otherwise, set existing ttl name
                    # when disabling
                    'AttributeName': ttl_attribute_name if ttl_enabled
                        else existing_ttl_attribute_name
                }
            )
            if wait:
                self._wait_for_table_update(table_name=table_name)
            return response

    def update_global_indexes(self, table_name, global_indexes_meta,
                              existing_global_indexes, table_read_capacity,
                              table_write_capacity, existing_capacity_mode):
        """ Creates, Deletes or Updates global indexes for the table

        :param table_name: DynamoDB table name
        :type table_name: str
        :param global_indexes_meta: list of global indexes definitions in meta
        :type global_indexes_meta: list
        :param existing_global_indexes: list of global indexes currently
            existing in table
        :type existing_global_indexes: list
        :param table_read_capacity: table read capacity defined in meta
        :type table_read_capacity: int
        :param table_write_capacity: table write capacity defined in meta
        :type table_write_capacity: int
        :param existing_capacity_mode: capacity mode currently set in the table
        :type existing_capacity_mode: str
        :returns None
        """
        gsi_names = [gsi.get('name') for gsi in global_indexes_meta]
        existing_gsi_names = [gsi.get('IndexName')
                              for gsi in existing_global_indexes]

        global_indexes_to_delete = []
        for gsi in existing_global_indexes:
            if gsi.get('IndexName') not in gsi_names:
                global_indexes_to_delete.append(gsi)

        global_indexes_to_create = []
        for gsi in global_indexes_meta:
            if gsi.get('name') not in existing_gsi_names:
                global_indexes_to_create.append(gsi)

        for gsi in global_indexes_to_delete:
            index_name = gsi.get('IndexName')
            self.delete_global_secondary_index(table_name=table_name,
                                               index_name=index_name)
            self._wait_for_index_update(table_name, index_name)
            _LOG.info(
                'Removed global secondary index {0} for table {1}'.format(
                    index_name, table_name))

        for gsi in global_indexes_to_create:
            read_capacity = \
                gsi.get('read_capacity') or table_read_capacity
            write_capacity = \
                gsi.get('write_capacity') or table_write_capacity
            self.create_global_secondary_index(
                table_name=table_name, index_meta=gsi,
                existing_capacity_mode=existing_capacity_mode,
                read_throughput=read_capacity, write_throughput=write_capacity
            )
            self._wait_for_index_update(table_name, gsi.get('name'))
            _LOG.info(
                'Created global secondary index {0} for table {1}'.format(
                    gsi.get('name'), table_name))

    def _wait_for_table_update(self, table_name, sleep_amount=20,
                               max_attempts=25):
        """ Waits for table to go into ACTIVE state.

        :param table_name: DynamoDB table name
        :type table_name: str
        :param sleep_amount: time in seconds to wait for the next attempt
        :type sleep_amount: int
        :param max_attempts: maximum attempts to ping the table
        :type max_attempts: int
        :returns None
        """
        table = self.get_table_by_name(table_name)
        table.wait_until_exists(
            WaiterConfig={
                'Delay': sleep_amount,
                'MaxAttempts': max_attempts
            }
        )

    def _wait_for_index_update(self, table_name, index_name, sleep_amount=20,
                               max_attempts=25):
        """ Waits for global secondary index
        to be able to continue table update.

        :param table_name: DynamoDB table name
        :type table_name: str
        :param index_name: name of the global index to wait for
        :type index_name: str
        :param sleep_amount: time in seconds to wait for the next attempt
        :type sleep_amount: int
        :param max_attempts: maximum attempts to ping the table
        :type max_attempts: int
        :returns None
        """
        num_attempts = 0
        while True:
            num_attempts += 1
            table = self.describe_table(table_name)
            indexes = table.get('GlobalSecondaryIndexes', [])
            index = next(
                (idx for idx in indexes if idx['IndexName'] == index_name),
                None
            )
            # there is no index which means it was successfully deleted or
            # the index is in backfilling state which allows
            # to continue table update
            if index is None or index.get('Backfilling'):
                return
            if num_attempts >= max_attempts:
                raise ResourceProcessingError('Max attempts exceeded')
            time.sleep(sleep_amount)

    def delete_global_secondary_index(self, table_name, index_name):
        """ Deletes global secondary index from the specified table.

        :param table_name: DynamoDB table name
        :type table_name: str
        :param index_name: name of the global index to delete
        :type index_name: str
        :returns update_table response as dict
        """
        response = self.client.update_table(
            TableName=table_name,
            GlobalSecondaryIndexUpdates=[
                {
                    'Delete': {
                        'IndexName': index_name
                    }
                }
            ]
        )
        return response

    def create_global_secondary_index(self, table_name, index_meta,
                                      existing_capacity_mode,
                                      read_throughput=None,
                                      write_throughput=None):
        """ Creates global secondary index for the specified table.

        :param table_name: DynamoDB table name
        :type table_name: str
        :param index_meta: global index info defined in meta
        :type index_meta: dict
        :param existing_capacity_mode: capacity mode currently set in the table
        :type existing_capacity_mode: str
        :param read_throughput: read capacity to assign for global index
        :type read_throughput: int
        :param write_throughput: write capacity to assign for global index
        :type write_throughput: int
        :returns update_table response as dict
        """
        index_info = _build_global_index_definition(
            index=index_meta,
            billing_mode=existing_capacity_mode,
            read_throughput=read_throughput,
            write_throughput=write_throughput
        )
        definitions = []
        _add_index_keys_to_definition(definition=definitions, index=index_meta)

        response = self.client.update_table(
            TableName=table_name,
            AttributeDefinitions=definitions,
            GlobalSecondaryIndexUpdates=[
                {
                    'Create': index_info
                }
            ]
        )
        return response

    def update_global_secondary_index(self, table_name, index_name,
                                      read_throughput=None,
                                      write_throughput=None):
        """ Updates global secondary index capacity for the specified table.

        :param table_name: DynamoDB table name
        :type table_name: str
        :param index_name: name of the index to update
        :type index_name: str
        :param read_throughput: read capacity to assign for global index
        :type read_throughput: int
        :param write_throughput: write capacity to assign for global index
        :type write_throughput: int
        :returns update_table response as dict
        """
        read_throughput = read_throughput or DEFAULT_GI_READ_CAPACITY
        write_throughput = write_throughput or DEFAULT_GI_WRITE_CAPACITY
        response = self.client.update_table(
            TableName=table_name,
            GlobalSecondaryIndexUpdates=[
                {
                    'Update': {
                        'IndexName': index_name,
                        'ProvisionedThroughput': {
                            'ReadCapacityUnits': read_throughput,
                            'WriteCapacityUnits': write_throughput
                        }
                    }
                }
            ]
        )
        return response

    def enable_table_stream(self, table_name,
                            stream_type='NEW_AND_OLD_IMAGES'):
        """ Sets StreamEnabled=true for the specified table.

        :type table_name: str
        :type stream_type: str
        """
        table = self.get_table_by_name(table_name)
        table.update(
            StreamSpecification={
                'StreamEnabled': True,
                'StreamViewType': stream_type
            }
        )

    def disable_table_stream(self, table_name,
                             stream_type='NEW_AND_OLD_IMAGES'):
        """ Sets StreamEnabled=false for the specified table.

        :type table_name: str
        :type stream_type: str
        """
        table = self.get_table_by_name(table_name)
        table.update(
            StreamSpecification={
                'StreamEnabled': False,
                'StreamViewType': stream_type
            }
        )

    def is_stream_enabled(self, table_name):
        table = self.get_table_by_name(table_name)
        stream = table.stream_specification
        if stream:
            return stream.get('StreamEnabled')

    def update_table_capacity(self, table_name, billing_mode,
                              read_capacity, write_capacity,
                              existing_billing_mode,
                              existing_provisioned_throughput,
                              existing_on_demand_throughput,
                              existing_global_indexes, wait=True):
        """ Updates table capacity configuration.

        :param table_name: DynamoDB table name
        :type table_name: str
        :param billing_mode: capacity mode to set for the table
        :type billing_mode: str
        :param read_capacity: read capacity to assign for the table
        :type read_capacity: int
        :param write_capacity: write capacity to assign for the table
        :type write_capacity: int
        :param existing_billing_mode: capacity mode currently set in the table
        :type billing_mode: str
        :param existing_provisioned_throughput: provisioned throughput
                                                currently set in the table
        :type existing_provisioned_throughput: dict
        :param existing_on_demand_throughput: on demand throughput currently
                                              set in the table
        :type existing_on_demand_throughput: dict
        :param existing_global_indexes: global secondary indexes already
            present in the table
        :type existing_global_indexes: list
        :param wait: to wait for table update to finish or not
        :type wait: bool
        :returns update_table response as boto3.DynamoDB.Table object or None
            if there were no changes made
        """
        params = {}

        global_secondary_indexes_updates = []

        if billing_mode != existing_billing_mode:
            if billing_mode == 'PROVISIONED':
                params['ProvisionedThroughput'] = dict(
                    ReadCapacityUnits=read_capacity or DEFAULT_READ_CAPACITY,
                    WriteCapacityUnits=write_capacity or DEFAULT_WRITE_CAPACITY
                )

                for gsi in existing_global_indexes:
                    ind_rcu = gsi.get('read_capacity')
                    ind_wcu = gsi.get('write_capacity')
                    global_secondary_indexes_updates.append({
                        'Update': {
                            'IndexName': gsi.get('IndexName'),
                            'ProvisionedThroughput': dict(
                                ReadCapacityUnits=
                                ind_rcu or read_capacity or
                                DEFAULT_GI_READ_CAPACITY,
                                WriteCapacityUnits=
                                ind_wcu or write_capacity or
                                DEFAULT_GI_WRITE_CAPACITY
                            )
                        }
                    })
            else:
                params['OnDemandThroughput'] = dict(
                    MaxReadRequestUnits=
                    read_capacity or UNLIMITED_REQUEST_UNIT,
                    MaxWriteRequestUnits=
                    write_capacity or UNLIMITED_REQUEST_UNIT
                )

                for gsi in existing_global_indexes:
                    ind_rcu = gsi.get('read_capacity')
                    ind_wcu = gsi.get('write_capacity')
                    global_secondary_indexes_updates.append({
                        'Update': {
                            'IndexName': gsi.get('IndexName'),
                            'OnDemandThroughput': dict(
                                MaxReadRequestUnits=
                                ind_rcu or read_capacity or
                                UNLIMITED_REQUEST_UNIT,
                                MaxWriteRequestUnits=
                                ind_wcu or write_capacity or
                                UNLIMITED_REQUEST_UNIT
                            )
                        }
                    })
        else:
            if billing_mode == 'PROVISIONED':
                if (read_capacity !=
                        existing_provisioned_throughput['ReadCapacityUnits']
                        or write_capacity !=
                        existing_provisioned_throughput['WriteCapacityUnits']):
                    params['ProvisionedThroughput'] = dict(
                        ReadCapacityUnits=
                        read_capacity or DEFAULT_READ_CAPACITY,
                        WriteCapacityUnits=
                        write_capacity or DEFAULT_WRITE_CAPACITY)

                    for gsi in existing_global_indexes:
                        ind_rcu = gsi.get('read_capacity')
                        ind_wcu = gsi.get('write_capacity')
                        global_secondary_indexes_updates.append({
                            'Update': {
                                'IndexName': gsi.get('IndexName'),
                                'ProvisionedThroughput': dict(
                                    ReadCapacityUnits=
                                    ind_rcu or read_capacity or
                                    DEFAULT_GI_READ_CAPACITY,
                                    WriteCapacityUnits=
                                    ind_wcu or write_capacity or
                                    DEFAULT_GI_WRITE_CAPACITY
                                )
                            }
                        })
            else:
                if (read_capacity !=
                        existing_on_demand_throughput['MaxReadRequestUnits']
                        or write_capacity !=
                        existing_on_demand_throughput['MaxWriteRequestUnits']):
                    params['OnDemandThroughput'] = dict(
                        MaxReadRequestUnits=
                        read_capacity or UNLIMITED_REQUEST_UNIT,
                        MaxWriteRequestUnits=
                        write_capacity or UNLIMITED_REQUEST_UNIT
                    )

                    for gsi in existing_global_indexes:
                        ind_rcu = gsi.get('read_capacity')
                        ind_wcu = gsi.get('write_capacity')
                        global_secondary_indexes_updates.append({
                            'Update': {
                                'IndexName': gsi.get('IndexName'),
                                'OnDemandThroughput': dict(
                                    MaxReadRequestUnits=
                                    ind_rcu or read_capacity or
                                    UNLIMITED_REQUEST_UNIT,
                                    MaxWriteRequestUnits=
                                    ind_wcu or write_capacity or
                                    UNLIMITED_REQUEST_UNIT
                                )
                            }
                        })

        if global_secondary_indexes_updates:
            params['GlobalSecondaryIndexUpdates'] = \
                global_secondary_indexes_updates

        if params:
            params.update(TableName=table_name,
                          BillingMode=billing_mode)
            _LOG.debug(f'Updating {table_name} table capacity. Table capacity '
                       f'mode: {billing_mode}, meta read/write '
                       f'capacities: {read_capacity}/{write_capacity}. '
                       f'Update params: {params}')
            self.client.update_table(**params)
            if wait:
                self._wait_for_table_update(table_name=table_name,
                                            sleep_amount=30,
                                            max_attempts=50)
            return self.get_table_by_name(table_name)

    def get_table_by_name(self, table_name):
        """ Get table by name.

        :type table_name: str
        :returns table
        """
        return self.conn.Table(table_name)

    def query_by_index(self, table_name, index_name, expression):
        table = self.get_table_by_name(table_name)
        return table.query(
            IndexName=index_name,
            KeyConditionExpression=expression
        )

    def query_by_index_and_expression(self, table_name, index_name, key,
                                      expression):
        table = self.get_table_by_name(table_name)
        return table.query(
            IndexName=index_name,
            KeyConditionExpression=key,
            FilterExpression=expression
        )

    def describe_table(self, table_name):
        try:
            return self.client.describe_table(TableName=table_name)['Table']
        except ClientError as e:
            if 'ResourceNotFoundException' in str(e):
                pass  # valid exception
            else:
                raise e

    def put_item(self, table_name, item):
        """ Creates a new item or updates existing in specified table.

        :type table_name: str
        :type item: dict
        """
        table = self.conn.Table(table_name)
        table.put_item(Item=item)

    def put_with_sort_by_date(self, resources_to_put, table_name):
        """ Saves billing records by date in ascending order

        :type resources_to_put: list
        :param resources_to_put: list with billing records

        :type table_name: str
        :param table_name: DynamoDB table name
        """
        for each in sorted(resources_to_put, key=itemgetter('d')):
            self.put_item(table_name, each)

    def items_batch_write(self, table_name, items):
        """ Write items to table from dict.

        :type table_name: str
        :type items: list
        """
        if items:
            table = self.conn.Table(table_name)
            with table.batch_writer() as batch:
                for each in items:
                    if isinstance(each, dict):
                        batch.put_item(Item=each)

    def items_batch_get(self, table_name, hash_key_name, hash_keys_values,
                        sort_key_name=None, sort_key_value=None):
        if hash_keys_values:
            hash_values_list = list()
            for hash_value in hash_keys_values:
                if sort_key_name and sort_key_value:
                    key = {
                        hash_key_name: hash_value,
                        sort_key_name: sort_key_value
                    }
                else:
                    key = {
                        hash_key_name: hash_value
                    }
                hash_values_list.append(key)
            results = []
            response = self.conn.batch_get_item(
                RequestItems={
                    table_name: {
                        'Keys': hash_values_list
                    }
                }
            )
            results.extend(response['Responses'][table_name])
            while response['UnprocessedKeys']:
                response = self.conn.batch_get_item(
                    RequestItems={
                        table_name: {
                            'Keys': hash_values_list
                        }
                    }
                )
                results.extend(response['Responses'][table_name])
            return results

    def get_item(self, table_name, hash_key_name, hash_key_value,
                 sort_key_name=None, sort_key_value=None):
        """ Get item from specified table by pk.

        :type table_name: str
        :type hash_key_name: str
        :param hash_key_value: str/number
        :type sort_key_name: str
        :param sort_key_value: str/number
        :returns list of items(if exists)
        """
        table = self.get_table_by_name(table_name)
        item = {hash_key_name: hash_key_value}
        if sort_key_name and sort_key_value:
            item[sort_key_name] = sort_key_value
        result = table.get_item(Key=item, ConsistentRead=True)
        if 'Item' in result:
            return result['Item']

    def scan(self, table_name=None, table=None, token=None,
             filter_expr=None, limit=None, attr_to_select='ALL_ATTRIBUTES'):
        """ DynamoDB table scan with consistent read and custom retry.

        :type table_name: str
        :type table: Table
        :type token: dict
        :type filter_expr: Attr
        :type limit: int
        :type attr_to_select: str
        """
        if table and table_name:
            raise ParameterError("'table' OR 'table_name' must be set.")
        elif table:
            pass  # just use Table object
        elif table_name:
            table = self.get_table_by_name(table_name)
        else:
            raise ParameterError("'table' or 'table_name' must be set.")

        params = {'ConsistentRead': True, 'Select': attr_to_select}
        if token:
            params['ExclusiveStartKey'] = token
        if filter_expr:
            params['FilterExpression'] = filter_expr
        if limit:
            params['Limit'] = limit

        return table.scan(**params)

    def get_all_items(self, table_name, filter_expr=None):
        """ Get all items from table.

        :type table_name: str
        :type filter_expr: Attr
        :return list(if items exist)
        """
        table = self.get_table_by_name(table_name)
        response = self.scan(table=table, filter_expr=filter_expr)
        do_continue = response.get('LastEvaluatedKey')
        items = []
        if 'Items' in response:
            items.extend(response['Items'])
        while do_continue:
            response = self.scan(table=table, token=do_continue,
                                 filter_expr=filter_expr)
            do_continue = response.get('LastEvaluatedKey')
            if 'Items' in response:
                items.extend(response['Items'])
        return items

    def _scan_all(self, table_name, func, limit=None, filter_expr=None, *args,
                  **kwargs):
        """ Calls a function for each item in the table.

        :type table_name: str
        :param func: Function to call for each item.
                     It must accept the item as the first argument.
        :type func: function
        :param args: any args the function takes after the item
        :param kwargs: any kwargs the func takes after the item and args
        """
        table = self.get_table_by_name(table_name)
        response = self.scan(table=table, filter_expr=filter_expr, limit=limit)
        do_continue = response.get('LastEvaluatedKey')
        if 'Items' in response:
            for each in response['Items']:
                func(each, *args, **kwargs)
        while do_continue:
            response = self.scan(table=table, token=do_continue, limit=limit,
                                 filter_expr=filter_expr)
            do_continue = response.get('LastEvaluatedKey')
            if 'Items' in response:
                for each in response['Items']:
                    func(each, *args, **kwargs)

    def for_each_item(self, table_name, func, *args, **kwargs):
        """ Go through all items in table and perform func.

        :type table_name: str
        :type func: function
        """
        self._scan_all(table_name, func, None, None, *args, **kwargs)

    def for_each_item_in_interval(self, start_interval, end_interval,
                                  interval_time, table_name, func,
                                  *args, **kwargs):
        """ Calls a function for each item in the table.

        :type start_interval: int (timestamp)
        :type end_interval: int (timestamp)
        :type interval_time: int (timestamp)
        :type table_name: str
        :param func: Function to call for each item.
                     It must accept the item as the first argument.
        :type func: function
        :param args: any args the function takes after the item
        :param kwargs: any kwargs the func takes after the item and args
        """
        table = self.get_table_by_name(table_name)
        start = start_interval
        count = 0
        while start < end_interval:
            response = self.scan(table=table,
                                 filter_expr=Attr('d').eq(start))
            do_continue = response.get('LastEvaluatedKey')
            if 'Items' in response:
                for each in response['Items']:
                    count += 1
                    func(each, *args, **kwargs)
            while do_continue:
                response = self.scan(table=table, token=do_continue,
                                     filter_expr=Attr('d').eq(start))
                do_continue = response.get('LastEvaluatedKey')
                if 'Items' in response:
                    for each in response['Items']:
                        count += 1
                        func(each, *args, **kwargs)
            start += interval_time

    def get_items_with_attribute_contains(self, table_name, attr_name, val):
        """ Get all items from table, which have specified attribute.

        :type table_name: str
        :type attr_name: str
        :type val: any
        :return list(if items exist)
        """
        return self.get_all_items(table_name, Attr(attr_name).contains(val))

    def get_items_with_attribute_value(self, table_name, attr_name,
                                       attr_value):
        """ Get all items from table, which have specified attribute with
        specified value.

        :type table_name: str
        :type attr_name: str
        :param attr_value: attr value in table
        :return: list(if items exist)
        """
        table = self.get_table_by_name(table_name)
        result = []
        response = self.scan(table=table,
                             filter_expr=Attr(attr_name).eq(attr_value))
        do_continue = response.get('LastEvaluatedKey')
        if 'Items' in response:
            result.extend(response['Items'])
        while do_continue:
            response = self.scan(table=table, token=do_continue,
                                 filter_expr=Attr(attr_name).eq(attr_value))
            do_continue = response.get('LastEvaluatedKey')
            if 'Items' in response:
                result.extend(response['Items'])
        return result

    def get_items_with_attr_between(self, table_name, attr_name,
                                    start, end):
        """ Get all items from table, which have specified attribute
        with specified value.

        :type table_name: str
        :type attr_name: str
        :param start: start attr value for range in table
        :param end: end attr value for range in table
        :return: list(if items exist)
        """
        return self.get_all_items(table_name,
                                  Attr(attr_name).between(start, end))

    def update_item(self, table_name, hash_key_name, hash_key_value,
                    key_to_set, value_to_set, sort_key_name=None,
                    sort_key_value=None):
        """ Updates some key in existing document using SET operator.

        :type table_name: str
        :type hash_key_name: str
        :param hash_key_value: value of the primary hash key
        :type key_to_set: str
        :param value_to_set: new value to set for the key_to_set
        :type sort_key_name: str
        :param sort_key_value: value of the primary sort key, if exists
        """
        self.flexible_update_item(table_name, hash_key_name, hash_key_value,
                                  'SET #k = :v', {'#k': key_to_set},
                                  {':v': value_to_set}, sort_key_name,
                                  sort_key_value)

    def flexible_update_item(self, table_name, hash_key_name, hash_key_value,
                             update_expression, expr_names, expr_values,
                             sort_key_name=None, sort_key_value=None):
        """ Updates some key in existing document using SET operator.

        :type table_name: str
        :param table_name: The name of the table.
        :type hash_key_name: str
        :param hash_key_name: name of the primary hash key
        :param hash_key_value: value of the primary hash key
        :type update_expression: str
        :param update_expression: Update expression with placeholders.
        :type expr_names: dict
        :param expr_names: expression attribute names
        :type expr_values: dict
        :param expr_values: expression attribute values
        :type sort_key_name: str
        :param sort_key_name: name of the primary sort key, if exists
        :param sort_key_value: value of the primary sort key, if exists
        """
        table = self.get_table_by_name(table_name)
        hash_key = {hash_key_name: hash_key_value}
        if sort_key_name and sort_key_value:
            hash_key[sort_key_name] = sort_key_value
        table.update_item(Key=hash_key, UpdateExpression=update_expression,
                          ExpressionAttributeNames=expr_names,
                          ExpressionAttributeValues=expr_values)

    def remove_specified_tables(self, table_names):
        """ Only 10 tables can be created, updated or deleted simultaneously.

        :type table_names: list of strings
        """
        waiters = {}
        start = 0
        end = 9
        while start < table_names:
            for name in table_names[:9]:
                table = self.get_table_by_name(name)
                waiters[table.name] = table.meta.client.get_waiter(
                    'table_not_exists')
                table.delete()
            for table_name in waiters:
                waiters[table_name].wait(TableName=table_name)

            start = end
            end += 9

    def table_exists(self, table_name):
        """ Check if table exists.

        :type table_name: str
        :return boolean
        """
        if table_name in self.get_tables_list():
            return True

    def get_table_stream_arn(self, table_name):
        """ Get table stream arn.

        :type table_name: str
        :return str(table stream name)
        """
        table = self.get_table_by_name(table_name)
        return table.latest_stream_arn

    def get_tables_list(self):
        """ Get all existing tables."""
        tables = []
        response = self.client.list_tables()
        token = response.get('LastEvaluatedTableName')
        tables.extend(response.get('TableNames'))
        while token:
            response = self.client.list_tables(ExclusiveStartTableName=token)
            token = response.get('LastEvaluatedTableName')
            tables.extend(response.get('TableNames'))
        return tables

    def remove_item(self, table_name, key_name, key_value, sort_name=None,
                    sort_value=None):
        """ Remove item from specified table by pk.

        :type table_name: str
        :type key_name: str
        :param key_value: value of attribute
        :type sort_name: str
        :param sort_value: value of attribute
        """
        table = self.get_table_by_name(table_name)
        key = {key_name: key_value}
        if sort_value and sort_name:
            key[sort_name] = sort_value
        table.delete_item(Key=key)

    def batch_remove_items(self, table_name, keys):
        table = self.get_table_by_name(table_name)
        with table.batch_writer() as batch:
            for key in keys:
                batch.delete_item(Key=key)

    def remove_table(self, table_name):
        """ Remove table and wait until AWS will perform operation.

        :type table_name: str
        """
        table = self.get_table_by_name(table_name)
        try:
            waiter = table.meta.client.get_waiter('table_not_exists')
            table.delete()
            waiter.wait(TableName=table_name)
        except ClientError as e:
            if e.response['Error']['Code'] == 'ResourceNotFoundException':
                _LOG.warn('Table %s is not found', table.name)
            else:
                raise e

    def remove_tables_by_names(self, table_names, log_not_found_error=True):
        """ Remove tables by names. AWS restricts simultaneous amount of
        tables, so after 10 requests sent we're waiting when operations finish
        and then continue.

        :type table_names: list
        :type log_not_found_error: boolean, parameter is needed for proper log
        handling in the retry decorator
        """
        removed_tables = []
        exceptions = []
        waiters = {}
        count = 0
        while count < len(table_names):
            tables_to_remove = table_names[count: count + 9]
            for name in tables_to_remove:
                table = self.get_table_by_name(name)
                waiters[table.name] = table.meta.client.get_waiter(
                    'table_not_exists')
                try:
                    table.delete()
                    removed_tables.append(name)
                except ClientError as e:
                    exception_type = e.response['Error']['Code']
                    if exception_type == 'ResourceNotFoundException':
                        _LOG.warn('Table %s is not found', table.name)
                        removed_tables.append(name)
                    else:
                        exceptions.append(str(e))
            count += 9
            for table_name in waiters:
                waiters[table_name].wait(TableName=table_name)
        return removed_tables, exceptions

    def _query(self, table_name=None, table=None, key_expr=None, token=None,
               limit=None, select='ALL_ATTRIBUTES'):
        if table and table_name:
            raise ParameterError("'table' OR 'table_name' must be set.")
        elif table:
            pass  # just use Table object
        elif table_name:
            table = self.get_table_by_name(table_name)
        else:
            raise ParameterError("'table' or 'table_name' must be set.")

        if not key_expr:
            raise ParameterError("'key_expr' must be set.")

        params = {
            'ConsistentRead': True, 'Select': select,
            'KeyConditionExpression': key_expr
        }
        if token:
            params['ExclusiveStartKey'] = token
        if limit:
            params['Limit'] = limit

        return table.query(**params)

    def query(self, table, key_expr, limit=1):
        """ Finds all documents with provided hash key value, sorts them
        in descending order and returns the first document. Example:
        For:
        {hk: hk, sk: 1}
        {hk: hk, sk: 2}
        {hk: hk, sk: 3}
        Will return:
        {hk: hk, sk: 3}

        :param table: table name
        :type table: str
        :type limit: int
        :type key_expr: Key
        :return:
        """
        table = self.get_table_by_name(table)
        items = []
        res = self._query(table=table, limit=limit, key_expr=key_expr)
        do_continue = res.get('LastEvaluatedKey')
        if 'Items' in res:
            items.extend(res['Items'])
        while do_continue:
            res = self._query(table=table, limit=limit, key_expr=key_expr,
                              token=do_continue)
            do_continue = res.get('LastEvaluatedKey')
            if 'Items' in res:
                items.extend(res['Items'])
        return items

    def query_by_hash_key(self, table, hash_name, hash_value, limit=1):
        key_expr = Key(hash_name).eq(hash_value)
        return self.query(table, key_expr, limit)



from syndicate.core.generators.deployment_resources.base_generator import \
    BaseDeploymentResourceGenerator
from syndicate.core.constants import STEP_FUNCTION_TYPE


class StepFunctionGenerator(BaseDeploymentResourceGenerator):
    RESOURCE_TYPE = STEP_FUNCTION_TYPE
    CONFIGURATION = {
        "definition": {
            "States": {
                "ExampleState": {
                    "Type": "Succeed"
                }
            },
            "Comment": "A description of your state machine",
            "StartAt": "ExampleState",
        },
        "event_sources": list,
        "dependencies": list,
        "iam_role": None,
        "tags": dict
    }



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import os
import shutil
from json import load
from typing import Any
from urllib.parse import urlparse

from syndicate.exceptions import ProjectStateError, \
    ResourceMetadataError, ResourceProcessingError, ParameterError, \
    InvalidValueError
from syndicate.commons.log_helper import get_logger, get_user_logger
from syndicate.core.build.helper import (build_py_package_name,
                                         resolve_bundle_directory)
from syndicate.core.build.validator.mapping import (VALIDATOR_BY_TYPE_MAPPING,
                                                    ALL_TYPES)
from syndicate.core.conf.processor import GLOBAL_AWS_SERVICES, \
    GLOBAL_AWS_SERVICE_PREFIXES
from syndicate.core.constants import (API_GATEWAY_TYPE, ARTIFACTS_FOLDER,
                                      BUILD_META_FILE_NAME, EBS_TYPE,
                                      LAMBDA_CONFIG_FILE_NAME, LAMBDA_TYPE,
                                      RESOURCES_FILE_NAME, RESOURCE_LIST,
                                      IAM_ROLE, LAMBDA_LAYER_TYPE,
                                      S3_PATH_NAME, APPSYNC_CONFIG_FILE_NAME,
                                      LAMBDA_LAYER_CONFIG_FILE_NAME,
                                      WEB_SOCKET_API_GATEWAY_TYPE,
                                      OAS_V3_FILE_NAME,
                                      API_GATEWAY_OAS_V3_TYPE, SWAGGER_UI_TYPE,
                                      SWAGGER_UI_CONFIG_FILE_NAME,
                                      TAGS_RESOURCE_TYPE_CONFIG,
                                      MVN_TARGET_DIR_NAME)
from syndicate.core.helper import (build_path, prettify_json,
                                   resolve_aliases_for_string,
                                   write_content_to_file, validate_tags)
from syndicate.core.resources.helper import resolve_dynamic_identifier, \
    detect_unresolved_aliases

DEFAULT_IAM_SUFFIX_LENGTH = 5
NAME_RESOLVING_BLACKLISTED_KEYS = [
    'prefix', 'suffix', 'resource_type', 'principal_service',
    'integration_type', 'authorization_type'
]

_LOG = get_logger(__name__)
USER_LOG = get_user_logger()


def validate_deployment_packages(bundle_path, meta_resources):
    package_paths = artifact_paths(meta_resources)
    nonexistent_packages = []
    for package in package_paths:
        package_path = build_path(bundle_path, package)
        if not os.path.exists(package_path):
            nonexistent_packages.append(package_path)

    if nonexistent_packages:
        raise ProjectStateError(
            f"Bundle is not properly configured. Nonexistent deployment "
            f"packages: '{prettify_json(nonexistent_packages)}'."
        )


def artifact_paths(meta_resources):
    return [i for i in
            [_retrieve_package(v) for v in list(meta_resources.values())] if i]


def _retrieve_package(meta):
    s3_path = meta.get(S3_PATH_NAME)
    if s3_path:
        return s3_path


def _check_duplicated_resources(initial_meta_dict, additional_item_name,
                                additional_item):
    """ Match two meta dicts (overall and separated) for duplicates.

    :type initial_meta_dict: dict
    :type additional_item_name: str
    :type additional_item: dict
    """
    if additional_item_name in initial_meta_dict:
        additional_type = additional_item['resource_type']
        initial_item = initial_meta_dict[additional_item_name]
        if not initial_item:
            return
        initial_type = initial_item['resource_type']
        if additional_type == initial_type and initial_type in \
                {API_GATEWAY_TYPE, WEB_SOCKET_API_GATEWAY_TYPE}:
            _LOG.info(
                f"The API Gateway '{additional_item_name}' is defined in "
                f"different deployment resources files. Going to merge "
                f"definitions..."
            )

            # return aggregated API description
            for param_name, initial_value in initial_item.items():
                if param_name == 'resource_type':
                    continue
                elif param_name in ['api_method_responses',
                                    'api_method_integration_responses',
                                    'cluster_cache_configuration',
                                    'cluster_throttling_configuration']:
                    additional_value = additional_item.get(param_name)
                    if initial_value and additional_value:
                        raise ResourceMetadataError(
                            f"Unable to merge the API Gateway "
                            f"'{additional_item_name}' definition because "
                            f"of duplication of the parameter '{param_name}' "
                            f"in different deployment resources files. "
                            f"Please resolve the conflict."
                        )
                    if initial_value:
                        additional_item[param_name] = initial_value

                elif param_name == 'dependencies':
                    dependencies_dict = {
                        each['resource_name']: each
                        for each in additional_item.get('dependencies') or []
                    }
                    if not additional_item.get('dependencies'):
                        additional_item['dependencies'] = []
                    for each in initial_value or []:
                        if each['resource_name'] not in dependencies_dict:
                            additional_item['dependencies'].append(each)

                elif param_name in ['binary_media_types', 'apply_changes']:
                    additional_item = _merge_api_gw_list_typed_configurations(
                        initial_item,
                        additional_item,
                        [param_name],
                        additional_item_name
                    )

                elif param_name in ['authorizers', 'tags', 'models',
                                    'resources']:
                    for each in list(initial_item.get(param_name, {}).keys()):
                        if each in list(
                                additional_item.get(param_name, {}).keys()):
                            raise ResourceMetadataError(
                                f"Unable to merge the API Gateway "
                                f"'{additional_item_name}' definition due to "
                                f"duplicate '{each}' key in the "
                                f"'{param_name}' property across different "
                                f"deployment resources files. "
                                f"Please resolve the conflict."
                            )

                    initial_param_value = initial_item.get(param_name) or {}
                    additional_param_value = additional_item.get(
                        param_name) or {}
                    additional_item[param_name] = {**initial_param_value,
                                                   **additional_param_value}

                elif additional_item.get(param_name):
                    raise ResourceMetadataError(
                        f"Unable to merge the API Gateway "
                        f"'{additional_item_name}' definition due to "
                        f"duplicate of the '{param_name}' property "
                        f"across different deployment resources files. "
                        f"Please resolve the conflict."
                    )

                else:
                    additional_item[param_name] = initial_value

            return additional_item

        else:
            raise ResourceProcessingError(
                f"Two resources with equal names were found! "
                f"Name: '{additional_item_name}', first resource type: "
                f"'{initial_type}', second resource type: "
                f"'{additional_type}'. \nPlease, rename one of them!"
            )


def _merge_api_gw_list_typed_configurations(initial_resource: dict,
                                            additional_resource: dict,
                                            property_names_list: list,
                                            additional_item_name: str):
    for property_name in property_names_list:
        initial_property_value = initial_resource.get(property_name, [])
        additional_resource_value = additional_resource.get(property_name, [])
        for each in initial_property_value:
            if each in additional_resource_value:
                raise ResourceMetadataError(
                    f"Unable to merge the API Gateway "
                    f"'{additional_item_name}' definition because "
                    f"of duplication of the parameter '{property_name}' "
                    f"in different deployment resources files. "
                    f"Please resolve the conflict."
                )

        additional_resource[
            property_name] = initial_property_value + additional_resource_value
    return additional_resource


def _populate_s3_path_python_node_dotnet(meta, bundle_name):
    name = meta.get('name')
    version = meta.get('version')
    prefix = meta.pop('prefix', None)
    suffix = meta.pop('suffix', None)
    if not name or not version:
        raise ParameterError(
            "Lambda config must contain name and version. Existing "
            f"configuration: '{prettify_json(meta)}'"
        )
    else:
        if prefix:
            name = name[len(prefix):]
        if suffix:
            name = name[:-len(suffix)]
        meta[S3_PATH_NAME] = build_path(bundle_name,
                                        build_py_package_name(name, version))


def _populate_s3_path_java(meta, bundle_name):
    deployment_package = meta.get('deployment_package')
    if not deployment_package:
        raise ParameterError(
            "Lambda config must contain deployment_package. Existing "
            f"configuration: '{prettify_json(meta)}'"
        )
    else:
        meta[S3_PATH_NAME] = build_path(bundle_name, deployment_package)


def _populate_s3_path_lambda(meta, bundle_name):
    runtime = meta.get('runtime')
    if not runtime:
        raise ParameterError(
            "Lambda config must contain runtime. "
            f"Existing configuration: '{prettify_json(meta)}'"
        )
    resolver_func = RUNTIME_PATH_RESOLVER.get(runtime.lower())
    if resolver_func:
        resolver_func(meta, bundle_name)
    else:
        raise InvalidValueError(
            f"Specified runtime '{runtime.lower()}' in '{meta.get('name')}' "
            f"is not supported. Supported runtimes: "
            f"'{list(RUNTIME_PATH_RESOLVER.keys())}'"
        )


def _populate_s3_path_lambda_layer(meta, bundle_name):
    deployment_package = meta.get('deployment_package')
    if not deployment_package:
        raise ParameterError(
            "Lambda Layer config must contain deployment_package. "
            f"'Existing configuration: '{prettify_json(meta)}'"
        )
    else:
        meta[S3_PATH_NAME] = build_path(bundle_name, deployment_package)


def _populate_s3_path_ebs(meta, bundle_name):
    deployment_package = meta.get('deployment_package')
    if not deployment_package:
        raise ParameterError(
            "Beanstalk_app config must contain deployment_package. "
            f"Existing configuration: '{prettify_json(meta)}'"
        )
    else:
        meta[S3_PATH_NAME] = build_path(bundle_name, deployment_package)


def _populate_s3_path_swagger_ui(meta, bundle_name):
    deployment_package = meta.get('deployment_package')
    if not deployment_package:
        raise ParameterError(
            "Swagger UI config must contain deployment_package. "
            f"Existing configuration: '{prettify_json(meta)}'"
        )
    else:
        meta[S3_PATH_NAME] = build_path(bundle_name, deployment_package)


def populate_s3_paths(overall_meta, bundle_name):
    for name, meta in overall_meta.items():
        resource_type = meta.get('resource_type')
        mapping_func = S3_PATH_MAPPING.get(resource_type)
        if mapping_func:
            mapping_func(meta, bundle_name)
    return overall_meta


def extract_deploy_stage_from_openapi_spec(openapi_spec: dict) -> str:
    """
    Extract the first path segment from the server URL in an API specification.
    If no server URL is found, or there is no path segment, raise an exception.
    """

    servers = openapi_spec.get('servers', [])
    if not servers:
        raise ParameterError(
            "No server information found in API specification."
        )

    server_url = servers[0].get('url', '')
    variables = servers[0].get('variables', {})

    # Substitute variables in the URL template with their default values,if any
    for var_name, var_details in variables.items():
        default_value = var_details.get('default', '')
        server_url = server_url.replace(f'{{{var_name}}}', default_value)

    # Extract the first path segment
    path_segments = [segment for segment in
                     urlparse(server_url).path.split('/')
                     if segment]
    if not path_segments:
        raise InvalidValueError("No path segments found in server URL.")

    return path_segments[0]


RUNTIME_PATH_RESOLVER = {
    'python3.9': _populate_s3_path_python_node_dotnet,
    'python3.10': _populate_s3_path_python_node_dotnet,
    'python3.11': _populate_s3_path_python_node_dotnet,
    'python3.12': _populate_s3_path_python_node_dotnet,
    'python3.13': _populate_s3_path_python_node_dotnet,
    'java11': _populate_s3_path_java,
    'java17': _populate_s3_path_java,
    'java21': _populate_s3_path_java,
    'nodejs18.x': _populate_s3_path_python_node_dotnet,
    'nodejs20.x': _populate_s3_path_python_node_dotnet,
    'nodejs22.x': _populate_s3_path_python_node_dotnet,
    'dotnet8': _populate_s3_path_python_node_dotnet
}

S3_PATH_MAPPING = {
    LAMBDA_TYPE: _populate_s3_path_lambda,
    EBS_TYPE: _populate_s3_path_ebs,
    LAMBDA_LAYER_TYPE: _populate_s3_path_lambda_layer,
    SWAGGER_UI_TYPE: _populate_s3_path_swagger_ui
}


def _look_for_configs(nested_files: list[str], resources_meta: dict[str, Any],
                      path: str, bundle_name: str) -> None:
    """ Look for all config files in project structure. Read content and add
    all meta to overall meta if there is no duplicates. If duplicates found -
    raise an exception.

    :param nested_files: A list of files in the project
    :param resources_meta: A dictionary of resources metadata
    :param path: A string path to the project
    :param bundle_name: A string name of the bundle
    """
    for each in nested_files:
        if each.endswith(LAMBDA_CONFIG_FILE_NAME) or \
                each.endswith(LAMBDA_LAYER_CONFIG_FILE_NAME) or \
                each.endswith(SWAGGER_UI_CONFIG_FILE_NAME) or \
                each.endswith(APPSYNC_CONFIG_FILE_NAME):
            resource_config_path = os.path.join(path, each)
            _LOG.debug(f'Processing file: {resource_config_path}')
            with open(resource_config_path) as data_file:
                resource_conf = load(data_file)

            resource_name = resource_conf['name']
            resource_type = resource_conf['resource_type']
            _LOG.debug(f'Found {resource_type}: {resource_name}')
            res = _check_duplicated_resources(resources_meta, resource_name,
                                              resource_conf)
            if res:
                resource_conf = res
            resources_meta[resource_name] = resource_conf

        if each.endswith(OAS_V3_FILE_NAME):
            openapi_spec_path = os.path.join(path, each)
            _LOG.debug(f'Processing file: {openapi_spec_path}')
            with open(openapi_spec_path) as data_file:
                openapi_spec = load(data_file)

            api_gateway_name = openapi_spec['info']['title']
            _LOG.debug(f'Found API Gateway: {api_gateway_name}')
            deploy_stage = extract_deploy_stage_from_openapi_spec(openapi_spec)
            resource = {
                "definition": openapi_spec,
                "resource_type": API_GATEWAY_OAS_V3_TYPE,
                "deploy_stage": deploy_stage,
            }
            tags = openapi_spec.get("x-syndicate-openapi-tags")
            if tags:
                resource["tags"] = tags
            res = _check_duplicated_resources(
                resources_meta, api_gateway_name, resource
            )
            if res:
                resource = res
            resources_meta[api_gateway_name] = resource

        if each == RESOURCES_FILE_NAME:
            additional_config_path = os.path.join(path, RESOURCES_FILE_NAME)
            _LOG.debug('Processing file: {0}'.format(additional_config_path))
            with open(additional_config_path, encoding='utf-8') as json_file:
                deployment_resources = load(json_file)
            for resource_name in deployment_resources:
                _LOG.debug('Found resource ' + resource_name)
                resource = deployment_resources[resource_name]
                # check if resource type exists in deployment framework and
                #  has resource_type field
                try:
                    resource_type = resource['resource_type']
                except KeyError:
                    error_message = (
                        f"There is no 'resource_type' in {resource_name} "
                        f"metadata"
                    )
                    _LOG.error(error_message)
                    raise ParameterError(error_message)
                if resource_type not in RESOURCE_LIST:
                    error_message = (
                        f'Unsupported resource type found: "{resource_type}". '
                        f'Please double-check the correctness of the specified '
                        f'resource type. To add a new resource type please '
                        f'request the support team.')
                    _LOG.error(error_message)
                    raise InvalidValueError(error_message)
                res = _check_duplicated_resources(resources_meta,
                                                  resource_name, resource)
                if res:
                    resource = res
                resources_meta[resource_name] = resource


# todo validate all required configs
def create_resource_json(project_path: str, bundle_name: str) -> dict[
    str, Any]:
    """ Create resource catalog json with all resource metadata in project.

    :param project_path: path to the project
    :type bundle_name: name of the bucket subdir
    """
    resources_meta = {}

    # Walking through every folder in the project
    for path, _, nested_items in os.walk(project_path):
        # there is no duplicates in single json, because json is a dict

        _look_for_configs(nested_items, resources_meta, path, bundle_name)

    meta_for_validation = _resolve_aliases(resources_meta)
    # check if all dependencies were described
    common_validator = VALIDATOR_BY_TYPE_MAPPING[ALL_TYPES]
    for name, meta in meta_for_validation.items():
        common_validator(resource_name=name,
                         resource_meta=meta, all_meta=meta_for_validation)

        resource_type = meta['resource_type']
        type_validator = VALIDATOR_BY_TYPE_MAPPING.get(resource_type)
        if type_validator:
            type_validator(name, meta)

    return resources_meta


def _resolve_names_in_meta(resources_dict, old_value, new_value):
    if isinstance(resources_dict, dict):
        for k, v in resources_dict.items():
            if k in NAME_RESOLVING_BLACKLISTED_KEYS:
                continue
            if isinstance(v, str) and old_value == v:
                resources_dict[k] = v.replace(old_value, new_value)
            elif isinstance(v, str) and old_value in v and v.startswith('arn'):
                resources_dict[k] = _resolve_name_in_arn(v, old_value, new_value)
            else:
                _resolve_names_in_meta(v, old_value, new_value)
    elif isinstance(resources_dict, list):
        for item in resources_dict:
            if isinstance(item, dict):
                _resolve_names_in_meta(item, old_value, new_value)
            elif (isinstance(item, str) and old_value in item and
                  item.startswith('arn')):
                index = resources_dict.index(item)
                resources_dict[index] = _resolve_name_in_arn(item, old_value, new_value)
            elif isinstance(item, str):
                if item == old_value:
                    index = resources_dict.index(old_value)
                    del resources_dict[index]
                    resources_dict.append(new_value)


def _resolve_name_in_arn(arn, old_value, new_value):
    from syndicate.core import CONFIG

    extended_prefix_mode = CONFIG.extended_prefix_mode
    arn_parts = arn.split(':')
    for part in arn_parts:
        new_part = None
        if part == old_value:
            new_part = new_value
        elif part.startswith(old_value) and part[len(old_value)] == '/':
            new_part = part.replace(old_value, new_value)
        elif part.endswith(old_value) and part[:-len(old_value)].endswith('/'):
            # to resolve resources with prefixes like ":role/", ":topic/", etc.
            resource_prefix = part[:-len(old_value)]
            if resource_prefix in GLOBAL_AWS_SERVICE_PREFIXES \
                    or extended_prefix_mode:
                new_part = part.replace(old_value, new_value)
        if new_part:
            index = arn_parts.index(part)
            arn_parts[index] = new_part
    return ':'.join(arn_parts)


def create_meta(project_path: str, bundle_name: str) -> None:
    # create overall meta.json with all resource meta info
    meta_path = build_path(project_path, ARTIFACTS_FOLDER,
                           bundle_name)
    _LOG.info(f'Bundle path: {meta_path}')
    overall_meta = create_resource_json(project_path=project_path,
                                        bundle_name=bundle_name)
    bundle_dir = resolve_bundle_directory(bundle_name=bundle_name)
    write_content_to_file(bundle_dir, BUILD_META_FILE_NAME, overall_meta)

    # remove Java runtime temporary files
    target_path = build_path(project_path, MVN_TARGET_DIR_NAME)
    if os.path.exists(target_path):
        _LOG.debug(f'Removing temporary directory \'{target_path}\'')
        shutil.rmtree(target_path, ignore_errors=True)


def resolve_meta(overall_meta):
    from syndicate.core import CONFIG
    iam_suffix = CONFIG.iam_suffix
    extended_prefix_mode = CONFIG.extended_prefix_mode
    overall_meta = _resolve_aliases(overall_meta)
    detect_unresolved_aliases(overall_meta)
    _LOG.debug('Resolved meta was created')
    _LOG.debug(prettify_json(overall_meta))
    _resolve_permissions_boundary(overall_meta)
    _LOG.debug('Permissions boundary were resolved')
    # get dict with resolved prefix and suffix in meta resources
    # key: current_name, value: resolved_name
    resolved_names = {}
    for name, res_meta in overall_meta.items():
        if res_meta.get('external'):
            continue
        resource_type = res_meta['resource_type']
        if resource_type in GLOBAL_AWS_SERVICES or extended_prefix_mode:
            resolved_name = resolve_resource_name(
                resource_name=name,
                prefix=CONFIG.resources_prefix,
                suffix=CONFIG.resources_suffix)
            if resource_type == LAMBDA_TYPE:
                res_meta['prefix'] = CONFIG.resources_prefix
                res_meta['suffix'] = CONFIG.resources_suffix
            # add iam_suffix to IAM role only if it is specified in config file
            if resource_type == IAM_ROLE and iam_suffix:
                resolved_name = resolved_name + iam_suffix
            if name != resolved_name:
                resolved_names[name] = resolved_name
    _LOG.debug('Going to resolve names in meta')
    _LOG.debug(f'Resolved names mapping: {str(resolved_names)}')
    for current_name, resolved_name in resolved_names.items():
        overall_meta[resolved_name] = overall_meta.pop(current_name)
        if not all([current_name, resolved_name]):
            continue
        _resolve_names_in_meta(overall_meta, current_name, resolved_name)
    return overall_meta


def resolve_tags(meta: dict) -> None:
    _LOG.debug('Going to resolve resources tags.')
    from syndicate.core import CONFIG
    common_tags = CONFIG.tags
    for res_name, res_meta in meta.items():
        res_tags = res_meta.get('tags', {})
        _LOG.debug(f'The resource {res_name} tags: {res_tags}')
        errors = validate_tags('tags', res_tags)
        if errors:
            USER_LOG.warn(
                f'The resource {res_name} tags don\'t pass validation and '
                f'will be removed from the resource meta. Details "{errors}"')
            res_meta.pop('tags')
            continue
        overall_tags = _format_tags(res_meta['resource_type'],
                                    {**common_tags, **res_tags})
        _LOG.debug(f'Resolved resource {res_name} tags {overall_tags}')
        res_meta['tags'] = overall_tags


def preprocess_tags(output: dict):
    for item in output.values():
        res_meta = item['resource_meta']
        tags = res_meta.get('tags')

        match tags:
            case tags if isinstance(tags, dict):
                continue
            case tags if isinstance(tags, list):
                res_meta['tags'] = _tags_to_dict(tags)
            case _:
                res_meta.pop('tags', None)


def _tags_to_dict(tags: list) -> dict:
    result = {}
    for tag in tags:
        tag_key = None
        tag_value = ''
        for k, v in tag.items():
            if k.lower() == 'key':
                tag_key = v
            if k.lower() == 'value':
                tag_value = v
        if tag_key is not None:
            result.update({tag_key: tag_value})
    return result


def _format_tags(res_type: str, tags: dict) -> dict | list:
    match res_type:
        case res_type if (res_type in
                          TAGS_RESOURCE_TYPE_CONFIG['capitalised_keys_list']):
            return [{'Key': k, 'Value': v} for k, v in tags.items()]
        case res_type if (res_type in
                          TAGS_RESOURCE_TYPE_CONFIG['lover_case_keys_list']):
            return [{'key': k, 'value': v} for k, v in tags.items()]
        case res_type if res_type in TAGS_RESOURCE_TYPE_CONFIG['untaggable']:
            _LOG.debug(f'The resource type {res_type} can not be tagged')
            return {}
        case _:
            return tags


def _resolve_aliases(overall_meta):
    """
    :type overall_meta: dict
    """
    from syndicate.core import CONFIG
    if CONFIG.aliases:
        aliases = {'${' + key + '}': str(value) for key, value in
                   CONFIG.aliases.items()}
        overall_meta = resolve_dynamic_identifier(aliases, overall_meta)
    return overall_meta


def _resolve_permissions_boundary(overall_meta):
    """Adds to every resource with resource_type IAM_ROLE permissions boundary
    from the config"""
    from syndicate.core import CONFIG
    if CONFIG.iam_permissions_boundary:
        for name, meta in overall_meta.items():
            if meta.get('resource_type') == IAM_ROLE:
                meta['permissions_boundary'] = CONFIG.iam_permissions_boundary


def resolve_resource_name(resource_name, prefix=None, suffix=None):
    return _resolve_suffix_name(
        _resolve_prefix_name(resource_name, prefix), suffix)


def resolve_resource_name_by_data(resource_name, resource_prefix,
                                  resource_suffix):
    return _resolve_suffix_name(
        _resolve_prefix_name(resource_name, resource_prefix), resource_suffix)


def _resolve_prefix_name(resource_name, resource_prefix):
    if resource_prefix:
        return resolve_aliases_for_string(resource_prefix) + resource_name
    return resource_name


def _resolve_suffix_name(resource_name, resource_suffix):
    if resource_suffix:
        return resource_name + resolve_aliases_for_string(resource_suffix)
    return resource_name


def get_meta_from_output(output: dict):
    from syndicate.core import CONFIG
    meta = {}
    for arn, data in output.items():
        resource_meta = data.get('resource_meta')
        resource_name = data.get('resource_name')

        suffix_index = resource_name.rfind(CONFIG.resources_suffix)
        if suffix_index != -1:  # if found
            resource_name = \
                resource_name[:suffix_index] + \
                resource_name[suffix_index + len(CONFIG.resources_suffix):]

        prefix_index = resource_name.find(CONFIG.resources_prefix)
        if prefix_index != -1:
            resource_name = \
                resource_name[:prefix_index] + \
                resource_name[prefix_index + len(CONFIG.resources_prefix):]

        meta.update({resource_name: resource_meta})

    return meta



from syndicate.core.generators.deployment_resources.base_generator import \
    BaseDeploymentResourceGenerator
from syndicate.core.constants import SQS_QUEUE_TYPE


class SQSQueueGenerator(BaseDeploymentResourceGenerator):
    RESOURCE_TYPE = SQS_QUEUE_TYPE
    CONFIGURATION = {
        "fifo_queue": bool,
        "region": None,
        "visibility_timeout": 30,  # seconds
        "delay_seconds": 0,
        "maximum_message_size": 1024,  # bytes
        "message_retention_period": 60,  # seconds
        "receive_message_wait_time_seconds": 0,
        "policy": dict,
        "redrive_policy": {
            "deadLetterTargetArn": None,
            "maxReceiveCount": None,
        },
        "kms_master_key_id": None,
        "kms_data_key_reuse_period_seconds": None,
        "content_based_deduplication": False,
        "tags": dict
    }

    def __init__(self, **kwargs):
        if 'dead_letter_target_arn' in kwargs:
            kwargs['deadLetterTargetArn'] = kwargs.pop(
                'dead_letter_target_arn')
            kwargs['maxReceiveCount'] = kwargs.pop('max_receive_count')
        super().__init__(**kwargs)



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from datetime import datetime, timezone

from syndicate.exceptions import InvalidValueError, ResourceNotFoundError
from syndicate.commons.log_helper import get_logger
from syndicate.core.helper import unpack_kwargs
from syndicate.core.helper import dict_keys_to_capitalized_camel_case
from syndicate.core.resources.base_resource import BaseResource
from syndicate.core.resources.helper import build_description_obj, \
    validate_params, assert_possible_values

_LOG = get_logger(__name__)

REQUIRED_PARAMS = {'name', 'schedule_expression', 'state', 'description',
                   'flexible_time_window'}


def convert_to_datetime(name, date_str):
    try:
        if len(date_str) > 10:
            return datetime.fromisoformat(date_str)
        else:
            return datetime.utcfromtimestamp(int(date_str))
    except (ValueError, OSError):
        raise InvalidValueError(
            f"Invalid date format: '{date_str}'. "
            f"Resource: '{name}'. Should be ISO8601 or timestamp"
            )


def prepare_schedule_parameters(meta):
    name = meta.get('name')
    validate_params(name, meta, REQUIRED_PARAMS)
    params = meta.copy()

    params = dict_keys_to_capitalized_camel_case(params)

    assert_possible_values([params.get('State')],
                           ['ENABLED', 'DISABLED']) \
        if 'State' in params else None
    assert_possible_values([params.get('FlexibleTimeWindow').get('Mode')],
                           ['OFF', 'FLEXIBLE']) \
        if 'Mode' in params.get('FlexibleTimeWindow') else None

    if 'StartDate' in params:
        start_date = convert_to_datetime(name, params.get('StartDate'))
        if start_date <= datetime.now(timezone.utc):
            raise InvalidValueError('Start date must be in the future.')
    if 'EndDate' in params:
        end_date = convert_to_datetime(name, params.get('EndDate'))
        if start_date <= datetime.now(timezone.utc):
            raise InvalidValueError('End date must be in the future.')

    if 'StartDate' in params and 'EndDate' in params:
        if start_date >= end_date:
            raise InvalidValueError(
                'Start date must be earlier than end date.'
            )

    return params


class EventBridgeSchedulerResource(BaseResource):

    def __init__(self, eventbridge_conn):
        self.connection = eventbridge_conn

    def create_schedule(self, args):
        return self.create_pool(self._create_schedule_from_meta, args)

    @unpack_kwargs
    def _create_schedule_from_meta(self, name, meta):
        _LOG.debug(f'Creating schedule {name}')
        check_params = meta['schedule_content']
        check_params['name'] = name
        params = prepare_schedule_parameters(check_params)
        response = self.describe_schedule(name, meta)
        if response:
            return response

        arn = self.connection.create_schedule(**params)
        _LOG.info(f'Created EventBridge schedule {arn}')
        return self.describe_schedule(name=name, meta=meta)

    def update_schedule(self, args):
        return self.create_pool(self._update_schedule_from_meta, args)

    @unpack_kwargs
    def _update_schedule_from_meta(self, name, meta, context):
        """ Update EventBridge Schedule from meta description after parameter
        validation.

        :type name: str
        :type meta: dict
        """
        check_params = meta['schedule_content']
        check_params['name'] = name
        group_name = check_params.get('group_name')
        response = self.connection.describe_schedule(name, meta)
        if not response:
            raise ResourceNotFoundError(f"'{name}' schedule does not exist.")

        params = prepare_schedule_parameters(check_params)

        arn = self.connection.update_schedule(**params)
        _LOG.info(f'Updated EventBridge schedule {arn}')
        return self.describe_schedule(name=name, meta=meta)

    def describe_schedule(self, name, meta):
        group_name = meta.get('group_name')

        response = self.connection.describe_schedule(name, group_name)
        if response:
            arn = response['Arn']
            return {
                arn: build_description_obj(response, name, meta)
            }
        return {}

    def remove_schedule(self, args):
        return self.create_pool(self._remove_schedule, args)

    @unpack_kwargs
    def _remove_schedule(self, arn, config):
        name = config['resource_name']
        try:
            group_name = config['resource_meta']['schedule_content'].get(
                'group_name')
        except:
            group_name = None
        self.connection.delete_schedule(name=name,
                                        group_name=group_name,
                                        log_not_found_error=False)
        return {arn: config}



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from boto3 import client
from syndicate.commons.log_helper import get_logger

_LOG = get_logger(__name__)


class ResourceGroupsTaggingAPIConnection:
    """Resource Groups Tagging API connection class."""

    def __init__(
            self,
            region=None,
            aws_access_key_id=None,
            aws_secret_access_key=None,
            aws_session_token=None,
    ):
        self.region = region
        self.aws_access_key_id = aws_access_key_id
        self.aws_secret_access_key = aws_secret_access_key
        self.aws_session_token = aws_session_token
        self.client = client('resourcegroupstaggingapi', region,
                             aws_access_key_id=aws_access_key_id,
                             aws_secret_access_key=aws_secret_access_key,
                             aws_session_token=aws_session_token)
        _LOG.debug('Opened new Resource Groups Tagging API connection.')

    def tag_resources(
            self,
            resources_arns: list,
            tags: dict,
    ) -> dict:
        params = dict(ResourceARNList=resources_arns, Tags=tags)
        response = self.client.tag_resources(**params)
        return response.get('FailedResourcesMap')

    def untag_resources(
            self,
            resources_arns: list,
            tag_keys: list,
    ) -> dict:
        params = dict(ResourceARNList=resources_arns, TagKeys=tag_keys)
        response = self.client.untag_resources(**params)
        return response.get('FailedResourcesMap')



from syndicate.core.generators.deployment_resources.base_generator import \
    BaseDeploymentResourceGenerator
from syndicate.core.constants import IAM_POLICY


class IAMPolicyGenerator(BaseDeploymentResourceGenerator):
    RESOURCE_TYPE = IAM_POLICY
    CONFIGURATION = {
        'policy_content': dict,
        'tags': dict
    }

    def _resolve_configuration(self, defaults_dict=None) -> dict:
        result = super()._resolve_configuration()
        if not result['policy_content']:
            example_resource = {
                "Version": "2012-10-17",
                "Statement": [
                    {
                        "Action": "*",
                        "Effect": "Deny",
                        "Resource": "*"
                    }
                ]
            }
            result['policy_content'].update(example_resource)
        return result



"""
    Copyright 2024 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import json
import os
import shutil
import zipfile

from syndicate.exceptions import ResourceMetadataError, \
    ArtifactAssemblingError
from syndicate.commons.log_helper import get_logger, get_user_logger
from syndicate.core.constants import SWAGGER_UI_SPEC_NAME_TEMPLATE, \
    SWAGGER_UI_ARTIFACT_NAME_TEMPLATE, SWAGGER_UI_CONFIG_FILE_NAME
from syndicate.core.helper import build_path

FILE_DEPLOYMENT_RESOURCES = 'deployment_resources.json'
INDEX_FILE_NAME = 'index.html'

_LOG = get_logger(__name__)
USER_LOG = get_user_logger()


def assemble_swagger_ui(project_path, bundles_dir, **kwargs):
    from syndicate.core import CONFIG
    path_to_project = CONFIG.project_path
    src_path = build_path(path_to_project, project_path)
    _LOG.info(f'Swagger UI sources are located by path: {src_path}')

    for root, dirs, files in os.walk(src_path):
        for item in dirs:
            _LOG.info(f'Going to process Swagger UI \'{item}\'')

            conf_file_path = build_path(src_path, item,
                                        SWAGGER_UI_CONFIG_FILE_NAME)
            if not os.path.isfile(conf_file_path):
                raise ResourceMetadataError(
                    f'\'{SWAGGER_UI_CONFIG_FILE_NAME}\' file not found for '
                    f'Swagger UI \'{item}\'.')

            index_file_path = build_path(src_path, item,
                                         INDEX_FILE_NAME)
            if not os.path.isfile(index_file_path):
                raise ArtifactAssemblingError(
                    f'\'{index_file_path}\' file not found for '
                    f'Swagger UI \'{item}\'.')

            with open(conf_file_path) as file:
                swagger_conf = json.load(file)
            spec_path = swagger_conf.get('path_to_spec') \
                if swagger_conf.get('path_to_spec') else ''
            if not os.path.isabs(spec_path):
                spec_path = build_path(path_to_project, spec_path)
                USER_LOG.info(f'Path to specification file resolved as '
                              f'\'{spec_path}\'')
            if not os.path.isfile(spec_path):
                raise ArtifactAssemblingError(
                    f'Specification file not found for Swagger UI '
                    f'\'{item}\' in the path {spec_path}.')

            artifact_name = SWAGGER_UI_ARTIFACT_NAME_TEMPLATE.format(name=item)
            spec_name = SWAGGER_UI_SPEC_NAME_TEMPLATE.format(name=item)
            zip_file_path = build_path(src_path, item, artifact_name)
            swagger_conf['deployment_package'] = artifact_name

            with open(conf_file_path, "w") as file:
                json.dump(swagger_conf, file)

            with zipfile.ZipFile(zip_file_path, 'w') as zipf:
                zipf.write(index_file_path, INDEX_FILE_NAME)
                zipf.write(spec_path, spec_name)
            shutil.move(zip_file_path, build_path(bundles_dir,
                                                  artifact_name))

            _LOG.info(f'Swagger UI {item} was processed successfully')



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import functools
import os
import shutil
import subprocess
import zipfile
from contextlib import closing
from datetime import datetime, date
from pathlib import PurePath, Path
from typing import Union

from syndicate.exceptions import InvalidValueError, InvalidTypeError, \
    ConfigurationError
from syndicate.commons.log_helper import get_logger, get_user_logger
from syndicate.core.constants import ARTIFACTS_FOLDER, CACHE_DIR
from syndicate.core.helper import build_path

_LOG = get_logger(__name__)
USER_LOG = get_user_logger()


def build_py_package_name(lambda_name, lambda_version):
    return '{0}-{1}.zip'.format(lambda_name, lambda_version)


def file_path_length_checker(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except FileNotFoundError as e:
            path_from_e = e.filename
            if len(path_from_e) < 260:
                raise e
            raise InvalidValueError(
                f"The path '{path_from_e}' has exceeded the 260-character "
                f"length (current length: '{len(path_from_e)}'). This may "
                f"have caused the FileNotFoundError. Please verify that the "
                f"file exists and consider shortening the path or increasing "
                f"the maximum path length allowed by the operating system."
            )
    return wrapper


@file_path_length_checker
def zip_dir(
        basedir: str,
        name: str,
        archive_subfolder: str = None,
) -> None:
    """ Compresses a directory into a zip file """
    assert os.path.isdir(basedir), \
        f"The specified base directory does not exist: {basedir}"

    with closing(zipfile.ZipFile(name, "w", zipfile.ZIP_DEFLATED)) as z:
        for root, dirs, files in os.walk(basedir, followlinks=True):
            archive_root = os.path.join(
                archive_subfolder, os.path.relpath(root, basedir)) \
                if archive_subfolder else os.path.relpath(root, basedir)
            for fn in files:
                absfn = os.path.normpath(os.path.join(root, fn))
                zfn = os.path.normpath(os.path.join(archive_root, fn))
                z.write(absfn, zfn)


def merge_zip_files(zip1_path: str, zip2_path: str, output_path: str,
                    output_subfolder:str=None):
    """
    Merge two ZIP files into a new ZIP file.
    Files from zip2_path overwrite those from zip1_path in case of name conflicts.
    """
    with zipfile.ZipFile(output_path, 'w') as output_zip:
        for zip_path in [zip1_path, zip2_path]:
            if not os.path.isfile(zip_path):
                _LOG.warning(
                    f"Zip file '{zip_path}' does not exist or is not file. "
                    f"Skipping.")
                continue
            with zipfile.ZipFile(zip_path, 'r') as input_zip:
                for file_info in input_zip.infolist():
                    data = input_zip.read(file_info.filename)
                    new_path = (
                        os.path.join(output_subfolder, file_info.filename) if
                        output_subfolder else file_info.filename
                    )

                    normalized_path = os.path.normpath(new_path)
                    final_path = normalized_path.replace(
                        os.sep,'/') if os.sep != '/' else normalized_path

                    output_zip.writestr(final_path, data)


def run_external_command(command: list):
    result = subprocess.run(
        command,
        capture_output=True,
        text=True)

    return result.returncode, result.stdout, result.stderr


def remove_dir(path: Union[str, Path]):
    removed = False
    while not removed:
        _LOG.info(f'Trying to remove "{path}"')
        try:
            shutil.rmtree(path)
            removed = True
        except Exception as e:
            removed = True
            _LOG.warn(f'An error "{e}" occurred while '
                      f'removing artifacts "{path}"')


def _json_serial(obj):
    """JSON serializer for objects not serializable by default json code"""

    if isinstance(obj, (datetime, date)):
        return obj.isoformat()
    raise InvalidTypeError(f"Type '{type(obj).__name__}' is not serializable")


def resolve_all_bundles_directory():
    from syndicate.core import CONF_PATH
    return build_path(CONF_PATH, ARTIFACTS_FOLDER)


def resolve_bundle_directory(bundle_name):
    return build_path(resolve_all_bundles_directory(), bundle_name)

def resolve_bundles_cache_directory():
    return build_path(resolve_all_bundles_directory(), CACHE_DIR)


def assert_bundle_bucket_exists() -> None:
    from syndicate.core import CONFIG, CONN
    if not CONN.s3().is_bucket_exists(CONFIG.deploy_target_bucket):
        raise ConfigurationError(
            f"Bundles bucket '{CONFIG.deploy_target_bucket}' does not exist. "
            f"Please use 'create_deploy_target_bucket' to create the bucket."
        )


def construct_deploy_s3_key_path(bundle_name: str, deploy_name: str,
                                 is_failed: bool = False) -> str:
    from syndicate.core import CONFIG
    file_name = f"{deploy_name}{'_failed' if is_failed else ''}.json"
    return PurePath(CONFIG.deploy_target_bucket_key_compound, bundle_name,
                    'outputs', file_name).as_posix()



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import traceback
from concurrent.futures import ThreadPoolExecutor, as_completed

from syndicate.commons.log_helper import get_logger, get_user_logger
from syndicate.connection import ResourceGroupsTaggingAPIConnection
from syndicate.core.constants import LAMBDA_TYPE, SWAGGER_UI_TYPE, \
    TAGS_RESOURCE_TYPE_CONFIG
from syndicate.core.resources.helper import chunks

_LOG = get_logger(__name__)
USER_LOG = get_user_logger()


class TagsApiResource:
    def __init__(self, connection: ResourceGroupsTaggingAPIConnection):
        self.connection = connection
        self.resource_type_to_preprocessor_mapping = {
            LAMBDA_TYPE: self._preprocess_lambda_arn
        }
        self.post_deploy_tagging_types = TAGS_RESOURCE_TYPE_CONFIG[
            'post_deploy_tagging']
        self.untaggable_resource_types = TAGS_RESOURCE_TYPE_CONFIG[
            'untaggable']
        from syndicate.core import CONFIG
        self.tags = CONFIG.tags

    @staticmethod
    def _preprocess_lambda_arn(arn: str, meta: dict):
        """Extracts lambda's arn without version and alias """
        lambda_name = meta['resource_name']
        processed_arn = arn[:arn.index(lambda_name) + len(lambda_name)]
        _LOG.debug(f'Processing lambda\'s arn from \'{arn}\' '
                   f'to \'{processed_arn}\'')
        lambda_arn_params = processed_arn.split(':')
        cw_lg_arn = (f'arn:aws:logs:{lambda_arn_params[3]}:'
                     f'{lambda_arn_params[4]}:log-group:/aws/lambda/'
                     f'{lambda_name}')
        _LOG.debug(f'For lambda \'{lambda_name}\' resolved CloudWatch logs '
                   f'group ARN \'{cw_lg_arn}\'')
        return [processed_arn, cw_lg_arn]

    def _extract_arns(self, output: dict) -> list:
        arns = []
        _LOG.info(f'Extracting and processing arns from output')
        for arn, meta in output.items():
            resource_type = meta['resource_meta']['resource_type']
            if resource_type in [SWAGGER_UI_TYPE]:
                continue
            processor_func = self.resource_type_to_preprocessor_mapping.get(
                resource_type)
            if processor_func:
                arn = processor_func(arn, meta)
            if isinstance(arn, list):
                arns.extend(arn)
            else:
                arns.append(arn)
        return arns

    def remove_tags(
            self,
            output: dict,
    ) -> dict:
        failed_untags = {}
        for tags, res_group in self._group_output_by_tags(output):
            arns = self._extract_arns(res_group)
            with ThreadPoolExecutor() as executor:
                futures = [
                    executor.submit(
                        self.connection.untag_resources,
                        batch,
                        list(tags.keys()),
                    ) for batch in chunks(arns, 20)
                ]
                for future in as_completed(futures):
                    failed = future.result() or {}
                    failed_untags.update(failed)
        if failed_untags:
            USER_LOG.warning(
                f"Can't remove tags from resources: {[*failed_untags]}"
            )
        else:
            _LOG.info(f'Tags were successfully removed from all resources')
        return failed_untags

    def safe_remove_tags(
            self,
            output: dict,
    ) -> bool:
        try:
            failed = self.remove_tags(output)
            error_message = [
                f"'{k}': {v['ErrorMessage']}" for k, v in failed.items()
            ]
            error_message = "\n".join(error_message)
        except Exception as e:
            error_message = str(e)
        if not error_message:
            return True
        USER_LOG.warning(
            f"The next error occurred during resources tagging\n{error_message}"
        )
        return False

    def apply_tags(
            self,
            output: dict,
    ) -> dict:
        failed_tags = {}
        for tags, res_group in self._group_output_by_tags(output):
            arns = self._extract_arns(res_group)
            with ThreadPoolExecutor() as executor:
                futures = [
                    executor.submit(self.connection.tag_resources, batch, tags)
                    for batch in chunks(arns, 20)
                ]
                for future in as_completed(futures):
                    failed = future.result() or {}
                    failed_tags.update(failed)
        if not failed_tags:
            _LOG.info('Tags were successfully applied')

        return failed_tags

    def safe_apply_tags(
            self,
            output: dict,
    ) -> bool:
        try:
            output = {k: v for k, v in output.items() if
                      v['resource_meta']['resource_type'] in
                      self.post_deploy_tagging_types}
            failed = self.apply_tags(output)
            error_message = [
                f"'{k}': {v['ErrorMessage']}" for k, v in failed.items()
            ]
            error_message = "\n".join(error_message)
        except Exception as e:
            error_message = str(e)
            _LOG.debug(traceback.format_exc())
        if not error_message:
            return True
        USER_LOG.error(
            f"The next error/s occurred during resources "
            f"tagging\n{error_message}"
        )
        return False

    def update_tags(
            self,
            old_output: dict,
            new_output: dict,
    ) -> dict:
        failed_updates = {}
        for arn, res_meta in new_output.items():
            arns = []
            base_arn = None
            res_type = res_meta['resource_meta']['resource_type']

            if res_type in self.untaggable_resource_types:
                continue

            preprocess_arn = \
                self.resource_type_to_preprocessor_mapping.get(res_type)
            if preprocess_arn:
                arn = preprocess_arn(arn, res_meta)

            if isinstance(arn, list):
                arns.extend(arn)
                base_arn = arn[0]
            else:
                arns.append(arn)

            if base_arn:
                for old_arn in old_output.keys():
                    if old_arn.startswith(base_arn):
                        old_res_meta = old_output[old_arn]
                        break
            else:
                old_res_meta = old_output.get(arn)
            old_res_tags = old_res_meta['resource_meta'].get('tags', {})
            new_res_tags = res_meta['resource_meta'].get('tags', {})

            to_tag = {k: v for k, v in new_res_tags.items()
                      if k not in old_res_tags.keys() or
                      v != old_res_tags.get(k)}

            to_untag = [
                k for k in old_res_tags.keys() if k not in new_res_tags.keys()
            ]

            if to_tag:
                failed_tag = self.connection.tag_resources(arns, to_tag)
                failed_updates.update(failed_tag or {})

            if to_untag:
                failed_untag = self.connection.untag_resources(arns, to_untag)
                failed_updates.update(failed_untag or {})
        if not failed_updates:
            _LOG.info('Tags were updated successfully')

        return failed_updates

    def safe_update_tags(
            self,
            old_output: dict,
            new_output: dict,
    ) -> bool:
        try:
            failed = self.update_tags(old_output, new_output)
            error_message = [
                f"'{k}': {v['ErrorMessage']}" for k, v in failed.items()
            ]
            error_message = "\n".join(error_message)
        except Exception as e:
            error_message = str(e)
            _LOG.debug(traceback.format_exc())
        if not error_message:
            return True
        USER_LOG.error(
            f"The next error/s occurred during updating resources "
            f"tags\n{error_message}"
        )
        return False

    @staticmethod
    def _group_output_by_tags(output: dict) -> list[tuple]:
        result = []
        for res_meta in output.values():
            if res_meta['resource_meta'].get('tags'):
                res_meta['resource_meta']['tags'] = dict(
                    sorted(res_meta['resource_meta']['tags'].items()))

        tags_list = [v['resource_meta']['tags'] for v in
                     output.values() if v['resource_meta'].get('tags')]
        tags_ids = set([frozenset(tags.items()) for tags in tags_list])

        for tags_id in tags_ids:
            for tags in tags_list:
                if tags_id == frozenset(tags.items()):
                    res_tags = tags
                    break
            res_group = {k: v for k, v in output.items() if
                         tags_id == frozenset(
                             v['resource_meta'].get('tags', {}).items())}
            result.append((res_tags, res_group))
        return result



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import os
import shutil

from syndicate.exceptions import EnvironmentError
from syndicate.commons.log_helper import get_logger
from syndicate.core.constants import MVN_TARGET_DIR_NAME
from syndicate.core.helper import build_path, execute_command_by_path, USER_LOG

_LOG = get_logger(__name__)

VALID_EXTENSIONS = ('.jar', '.war', '.zip')


def assemble_java_mvn_lambdas(project_path: str, bundles_dir: str,
                              errors_allowed: bool = False,
                              skip_tests: bool = False, **kwargs):
    from syndicate.core import CONFIG

    _check_maven_is_installed()
    target_path = os.path.join(CONFIG.project_path, MVN_TARGET_DIR_NAME)
    src_path = build_path(CONFIG.project_path, project_path)
    _LOG.info(f'Java sources are located by path: {src_path}')
    _LOG.info(f'Going to process java mvn project by path: '
              f'{CONFIG.project_path}')

    mvn_path = shutil.which('mvn')
    if mvn_path is None:
        raise EnvironmentError(
            'It seems that Apache Maven is not installed. Therefore, Java '
            'artifacts cannot be assembled. Please make sure that Apache '
            'Maven is installed.'
        )

    command = [mvn_path, 'clean', 'install']

    if skip_tests:
        command.append('-DskipTests')

    if errors_allowed:
        command.append('-DerrorsAllowed')

    execute_command_by_path(
        command=command, path=CONFIG.project_path, shell=False)

    # copy java artifacts to the target folder
    for root, dirs, files in os.walk(target_path):
        for file in _filter_bundle_files(files):
            shutil.copyfile(build_path(root, file),
                            build_path(bundles_dir, file))
    _LOG.info('Java mvn project was processed successfully')


def _filter_bundle_files(files: list[str]) -> list[str]:
    filtered_files = []
    exclude_prefix = 'original-'

    # to exclude redundant original-<lambda_name>.jar file
    # but do not exclude lambda jar if its name starts with 'original-'
    if sum(1 for item in files if item.startswith('original-')) > 1:
        exclude_prefix = 'original-original-'

    for file in files:
        if file.endswith(VALID_EXTENSIONS) and not \
                (file.startswith(exclude_prefix) and file.endswith('.jar')):
            filtered_files.append(file)
    return filtered_files


def _check_maven_is_installed():
    try:
        path = shutil.which('mvn')
        if not path:
            raise Exception
    except Exception:
        USER_LOG.error(
            'It seems like the Maven is not installed. There is no '
            'ability to build a Java bundle. Please, make sure Maven '
            'is installed and retry to build a bundle.')
        raise



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import os
from configparser import ConfigParser
from datetime import datetime
from typing import Union

import yaml

from syndicate.exceptions import ConfigurationError, \
    InvalidTypeError
from syndicate.commons.log_helper import get_logger
from syndicate.core.conf.bucket_view import \
    AbstractBucketView, AbstractViewDigest
from syndicate.core.conf.validator import \
    (PROJECT_PATH_CFG, REGION_CFG, DEPLOY_TARGET_BUCKET_CFG,
     ACCOUNT_ID_CFG, PROJECTS_MAPPING_CFG, AWS_ACCESS_KEY_ID_CFG,
     RESOURCES_PREFIX_CFG, RESOURCES_SUFFIX_CFG, AWS_SECRET_ACCESS_KEY_CFG,
     ConfigValidator, USE_TEMP_CREDS_CFG, SERIAL_NUMBER_CFG,
     TEMP_AWS_ACCESS_KEY_ID_CFG, TEMP_AWS_SECRET_ACCESS_KEY_CFG,
     TEMP_AWS_SESSION_TOKEN_CFG, EXPIRATION_CFG, TAGS_CFG,
     IAM_PERMISSIONS_BOUNDARY_CFG, LAMBDAS_ALIASES_NAME_CFG,
     AWS_SESSION_TOKEN_CFG, EXTENDED_PREFIX_MODE_CFG,
     LOCK_LIFETIME_MINUTES_CFG)
from syndicate.core.constants import (DEFAULT_SEP, IAM_POLICY, IAM_ROLE,
                                      S3_BUCKET_TYPE)

CONFIG_FILE_NAME = 'syndicate.yml'
ALIASES_FILE_NAME = 'syndicate_aliases.yml'

LEGACY_CONFIG_FILE_NAME = 'sdct.conf'
LEGACY_ALIASES_FILE_NAME = 'sdct_aliases.conf'

_LOG = get_logger(__name__)

GLOBAL_AWS_SERVICES = {IAM_ROLE, IAM_POLICY, S3_BUCKET_TYPE}
GLOBAL_AWS_SERVICE_PREFIXES = {'role/', 'policy/', 'user/', 'group/'}

DEFAULT_LOCK_TIME_IN_MINUTES = 20


class ConfigHolder:
    def __init__(self, dir_path):
        con_path_yml = os.path.join(dir_path, CONFIG_FILE_NAME)
        con_path_yaml = os.path.join(dir_path,
                                     CONFIG_FILE_NAME.replace('yml', 'yaml'))
        con_path = con_path_yml if \
            os.path.exists(con_path_yml) else con_path_yaml
        self._config_path = con_path

        conf_con_path = os.path.join(dir_path, LEGACY_CONFIG_FILE_NAME)

        if os.path.isfile(con_path):
            self._init_yaml_config(dir_path=dir_path, con_path=con_path)
        elif os.path.isfile(conf_con_path):
            self._init_conf_config(dir_path=dir_path)
        else:
            raise ConfigurationError(
                f"The syndicate configuration file is not found in the "
                f"directory '{dir_path}'"
            )

    def _assert_no_errors(self, errors: list):
        if errors:
            raise ConfigurationError(
                f"The following error occurred while '{self._config_path}' "
                f"parsing: '{errors}'"
            )

    def _init_yaml_config(self, dir_path, con_path):
        config_content = load_yaml_file_content(file_path=con_path)
        if config_content:
            validator = ConfigValidator(config_content)
            errors = validator.validate()
            self._assert_no_errors(errors)

        self._config_dict = config_content

        aliases_path_yml = os.path.join(dir_path, ALIASES_FILE_NAME)
        aliases_path_yaml = os.path.join(
            dir_path, ALIASES_FILE_NAME.replace('yml', 'yaml'))
        aliases_path = aliases_path_yml \
            if os.path.exists(aliases_path_yml) else aliases_path_yaml
        aliases_content = load_yaml_file_content(file_path=aliases_path)
        self._aliases = aliases_content
        self._aliases.update(self.default_aliases)

    def _init_conf_config(self, dir_path):
        con_path = os.path.join(dir_path, LEGACY_CONFIG_FILE_NAME)

        self._config_path = con_path
        self._config_dict = load_conf_file_content(self._config_path)

        validator = ConfigValidator(self._config_dict)
        errors = validator.validate()
        self._assert_no_errors(errors)

        alias_path = os.path.join(dir_path, LEGACY_ALIASES_FILE_NAME)
        if not os.path.exists(alias_path):
            _LOG.warn('sdct_aliases.conf does not exist '
                      'inside %s folder' % dir_path)

        self._aliases = load_conf_file_content(alias_path)
        self._aliases.update(self.default_aliases)

    def set_temp_credentials_to_config(self, temp_aws_access_key_id,
                                       temp_aws_secret_access_key,
                                       temp_aws_session_token,
                                       expiration):
        content_to_update = {
            TEMP_AWS_ACCESS_KEY_ID_CFG: temp_aws_access_key_id,
            TEMP_AWS_SECRET_ACCESS_KEY_CFG: temp_aws_secret_access_key,
            TEMP_AWS_SESSION_TOKEN_CFG: temp_aws_session_token,
            EXPIRATION_CFG: expiration
        }
        update_file_content(
            file_path=self._config_path,
            content=content_to_update
        )

    def _resolve_variable(self, variable_name):
        return self._config_dict.get(variable_name)

    def _resolve_aliases(self, variable_name):
        return self._aliases.get(variable_name)

    def _prepare_bucket_view(self) -> Union[None, AbstractBucketView]:
        """
        Prepares assigned bucket view instance,
        by providing the raw config payload.
        Under circumstances of an error, deletes the previously installed view,
        which defaults to using the raw format.
        :return: [None, AbstractBucketView]
        """
        view = self.deploy_target_bucket_view
        raw = self._resolve_variable(DEPLOY_TARGET_BUCKET_CFG)
        try:
            view.raw = raw
            _LOG.info(f'Viewing complement, {view.__class__.__name__},'
                      ' has been found, setting up the raw data.')
            return view

        except AttributeError:
            _LOG.warn('No viewing complement has been found.')
        except AbstractBucketView.BucketViewRuntimeError:
            _LOG.warn('Viewing complement set-up has failed.')

        del self.deploy_target_bucket_view
        return None

    def _resolve_bucket_view_attribute(self, attribute_name: str,
                                       default=None):
        """
        Retrieves bucket view value respectively to a provided attribute name.
        """
        if not isinstance(attribute_name, str):
            raise InvalidTypeError('Name of an attribute must be a string.')
        view = self.deploy_target_bucket_view
        if view and not view.raw:
            view = self._prepare_bucket_view()
        return getattr(view, attribute_name, default)

    @property
    def default_aliases(self):
        return {
            ACCOUNT_ID_CFG: self.account_id,
            REGION_CFG: self.region
        }

    @property
    def project_path(self):
        return path_resolver(self._resolve_variable(PROJECT_PATH_CFG))

    @property
    def account_id(self):
        return str(self._resolve_variable(ACCOUNT_ID_CFG))

    @property
    def lambdas_alias_name(self):
        return str(self._resolve_aliases(LAMBDAS_ALIASES_NAME_CFG))

    @property
    def access_role(self):
        return self._resolve_variable('access_role')

    @property
    def session_duration(self):
        duration = self._resolve_variable('session_duration')
        if duration:
            return int(self._resolve_variable('session_duration'))

    @property
    def aws_access_key_id(self):
        return self._resolve_variable(AWS_ACCESS_KEY_ID_CFG)

    @property
    def aws_session_token(self):
        return self._resolve_variable(AWS_SESSION_TOKEN_CFG)

    @property
    def aws_secret_access_key(self):
        return self._resolve_variable(AWS_SECRET_ACCESS_KEY_CFG)

    @property
    def region(self):
        return self._resolve_variable(REGION_CFG)

    @property
    def deploy_target_bucket(self) -> str:
        return self._resolve_bucket_view_attribute(
            'name', self._resolve_variable(DEPLOY_TARGET_BUCKET_CFG)
        )

    @property
    def deploy_target_bucket_key_compound(self) -> str:
        return self._resolve_bucket_view_attribute('key', '')

    @property
    def deploy_target_bucket_view(self) -> Union[AbstractBucketView, None]:
        return getattr(self, '_deploy_target_bucket_view', None)

    @deploy_target_bucket_view.setter
    def deploy_target_bucket_view(self, view: AbstractBucketView):
        if not isinstance(view, AbstractBucketView):
            _LOG.error('Bucket view couldn\'t have been set, '
                       'due to improper type.')
        elif not isinstance(view.digest, AbstractViewDigest):
            _LOG.error('Bucket view couldn\'t have been set,'
                       ' due to unassigned digest-parser property.')
        else:
            setattr(self, '_deploy_target_bucket_view', view)

    @deploy_target_bucket_view.deleter
    def deploy_target_bucket_view(self):
        delattr(self, '_deploy_target_bucket_view')

    @property
    def iam_permissions_boundary(self):
        return self._resolve_variable(IAM_PERMISSIONS_BOUNDARY_CFG)

    # mapping build tool : paths to project
    @property
    def build_projects_mapping(self):
        mapping_value = self._resolve_variable(PROJECTS_MAPPING_CFG)
        if type(mapping_value) == dict:
            return mapping_value
        if mapping_value:
            mapping_dict = {}
            for i in mapping_value.split(';'):
                key = i.split(':')[0]
                value = i.split(':')[1]
                list_values = mapping_dict.get(key)
                if list_values:
                    list_values.append(path_resolver(value))
                else:
                    mapping_dict[key] = [path_resolver(value)]
            return mapping_dict

    @property
    def resources_prefix(self):
        prefix = self._resolve_variable(RESOURCES_PREFIX_CFG)
        if prefix is None:
            return ''
        else:
            return prefix

    @property
    def resources_suffix(self):
        suffix = self._resolve_variable(RESOURCES_SUFFIX_CFG)
        if suffix is None:
            return ''
        else:
            return suffix

    @property
    def iam_suffix(self):
        """
        Optional property. It will be included as a ending
        of names for iam_roles.
        :return:
        """
        return self._resolve_variable('iam_suffix')

    @property
    def extended_prefix_mode(self):
        prefix_mode = self._resolve_variable(EXTENDED_PREFIX_MODE_CFG)
        return self._resolve_bool_param(prefix_mode)

    @property
    def aliases(self):
        return self._aliases

    @property
    def use_temp_creds(self):
        var = self._resolve_variable(USE_TEMP_CREDS_CFG)
        return self._resolve_bool_param(var)

    @property
    def serial_number(self):
        return self._resolve_variable(SERIAL_NUMBER_CFG)

    @property
    def temp_aws_access_key_id(self):
        return self._resolve_variable(TEMP_AWS_ACCESS_KEY_ID_CFG)

    @property
    def temp_aws_secret_access_key(self):
        return self._resolve_variable(TEMP_AWS_SECRET_ACCESS_KEY_CFG)

    @property
    def temp_aws_session_token(self):
        return self._resolve_variable(TEMP_AWS_SESSION_TOKEN_CFG)

    @property
    def expiration(self):
        return self._resolve_variable(EXPIRATION_CFG)

    @property
    def tags(self) -> dict:
        tags = self._resolve_variable(TAGS_CFG) or {}
        tags = {k: str(v) for k, v in tags.items()}
        return tags

    @property
    def lock_lifetime_minutes(self) -> int:
        return self._resolve_variable(LOCK_LIFETIME_MINUTES_CFG) or \
            DEFAULT_LOCK_TIME_IN_MINUTES

    def resolve_alias(self, name):
        if self._aliases.get(name):
            return self._aliases[name]

    @staticmethod
    def _resolve_bool_param(parameter):
        if isinstance(parameter, bool):
            return parameter
        elif isinstance(parameter, str):
            return parameter.lower() in ("yes", "true", "t", "1")
        return False


def path_resolver(path):
    return path.replace('\\', DEFAULT_SEP).replace('//', DEFAULT_SEP)


def str_to_bool(val):
    if isinstance(val, str):
        if val.lower() == 'true':
            return True
        elif val.lower() == 'false':
            return False
    return val


def str_to_datetime(val):
    if isinstance(val, str):
        if ' ' in val:
            val = val.replace(' ', 'T')
        return datetime.fromisoformat(val)
    return val


def add_default_section(file_path):
    with open(file_path, 'r+') as f:
        lines = f.readlines()
        lines = [line for line in lines if line.strip() != '']

        first_line = lines[0]
        f.seek(0)
        if '[default]' not in first_line:
            rest_of_file = f.read()
            f.seek(0)
            f.write('[default]\n' + rest_of_file)


def load_yaml_file_content(file_path):
    if not os.path.isfile(file_path):
        raise ConfigurationError(
            f"There is no file by path: '{file_path}'"
        )
    with open(file_path, 'r') as yaml_file:
        return yaml.load(yaml_file, Loader=yaml.FullLoader)


def load_conf_file_content(file_path):
    if not os.path.isfile(file_path):
        raise ConfigurationError(
            f"There is no file by path: '{file_path}'"
        )

    add_default_section(file_path)

    config = ConfigParser()
    config.read(file_path)
    config_dict = {}

    for section in config.sections():
        if section == 'tags':
            config_dict[section] = dict(config[section])
        else:
            section_dict = {
                k: str_to_bool(v) if k != 'expiration' else str_to_datetime(v)
                for k, v in dict(config[section]).items()}
            config_dict.update(section_dict)

    return config_dict


def update_file_content(file_path, content):
    if file_path.endswith('.yaml') or file_path.endswith('.yml'):
        update_yaml_file_content(file_path=file_path, content=content)
    elif file_path.endswith('.conf'):
        update_conf_file_content(file_path=file_path, content=content)


def update_yaml_file_content(file_path, content):
    file_content = load_yaml_file_content(file_path=file_path)
    file_content.update(content)
    with open(file_path, 'w') as yaml_file:
        yaml.dump(file_content, yaml_file, default_flow_style=False)


def update_conf_file_content(file_path, content):
    file_content = load_conf_file_content(file_path)
    file_content.update(content)

    config = ConfigParser()
    for key, val in file_content.items():
        if type(val) is dict:
            config.add_section(key)
            for sub_key, sub_val in val.items():
                config.set(key, sub_key, str(sub_val))
        else:
            if not config.has_section('default'):
                config.add_section('default')
            config.set('default', key, str(val))

    with open(file_path, 'w') as f:
        config.write(f)



from syndicate.core.generators.deployment_resources.\
    cloudwatch_event_rule_generator import CloudwatchEventRuleGenerator
from syndicate.core.constants import EVENT_BRIDGE_RULE_TYPE


class EventBridgeRuleGenerator(CloudwatchEventRuleGenerator):
    RESOURCE_TYPE = EVENT_BRIDGE_RULE_TYPE



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import concurrent
import glob
import json
import os
import platform
import shutil
import subprocess
import sys
from concurrent.futures import FIRST_EXCEPTION
from concurrent.futures import FIRST_EXCEPTION
from concurrent.futures.thread import ThreadPoolExecutor
from itertools import chain
from pathlib import Path
from typing import Union, Optional, List, Set

from syndicate.exceptions import ArtifactAssemblingError, InvalidTypeError
from syndicate.commons.log_helper import get_logger, get_user_logger
from syndicate.core.build.helper import build_py_package_name, zip_dir, \
    remove_dir, resolve_bundles_cache_directory, merge_zip_files
from syndicate.core.conf.processor import path_resolver
from syndicate.core.constants import (LAMBDA_CONFIG_FILE_NAME, DEFAULT_SEP,
                                      REQ_FILE_NAME, LOCAL_REQ_FILE_NAME,
                                      LAMBDA_LAYER_CONFIG_FILE_NAME,
                                      PYTHON_LAMBDA_LAYER_PATH,
                                      MANY_LINUX_2014_PLATFORM)
from syndicate.core.helper import (build_path, unpack_kwargs, zip_ext,
                                   without_zip_ext, compute_file_hash)
from syndicate.core.resources.helper import validate_params

_LOG = get_logger(__name__)
USER_LOG = get_user_logger()

_PY_EXT = "*.py"
EMPTY_LINE_CHARS = ('\n', '\r\n', '\t')

REQ_HASH_SUFFIX = 'r_hash'
EMPTY_FILE_HASH = 'e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855'
TMP_DIR = 'tmp'


def assemble_python_lambdas(project_path, bundles_dir, errors_allowed,
                            **kwargs):
    from syndicate.core import CONFIG
    project_base_folder = os.path.basename(os.path.normpath(project_path))
    if project_path != '.':
        project_abs_path = build_path(CONFIG.project_path, project_base_folder)
    else:
        project_abs_path = CONFIG.project_path
    _LOG.info('Going to process python project by path: {0}'.format(
        project_abs_path))
    with ThreadPoolExecutor(max_workers=5) as executor:
        futures = []
        for root, sub_dirs, files in os.walk(project_abs_path):
            for item in files:
                if item.endswith(LAMBDA_CONFIG_FILE_NAME):
                    _LOG.info('Going to build artifact in: {0}'.format(root))
                    arg = {
                        'root': str(Path(root)),
                        'config_file': str(Path(root, item)),
                        'target_folder': bundles_dir,
                        'project_path': project_path,
                        'errors_allowed': errors_allowed
                    }
                    futures.append(
                        executor.submit(_build_python_artifact, arg))
                elif item.endswith(LAMBDA_LAYER_CONFIG_FILE_NAME):
                    _LOG.info(f'Going to build lambda layer in `{root}`')
                    arg = {
                        'layer_root': root,
                        'bundle_dir': bundles_dir,
                        'project_path': project_path,
                        'errors_allowed': errors_allowed
                    }
                    futures.append(
                        executor.submit(build_python_lambda_layer, arg))
        result = concurrent.futures.wait(futures, return_when=FIRST_EXCEPTION)
    for future in result.done:
        exception = future.exception()
        if exception:
            raise ArtifactAssemblingError(exception)
    _LOG.info('Python project was processed successfully')


@unpack_kwargs
def build_python_lambda_layer(layer_root: str, bundle_dir: str,
                              project_path: str, errors_allowed: bool):
    """
    Layer root is a dir where these files exist:
    - lambda_layer_config.json
    - local_requirements.txt
    - requirements.txt
    """
    cache_dir_path = resolve_bundles_cache_directory()
    os.makedirs(cache_dir_path, exist_ok=True)

    with open(Path(layer_root, LAMBDA_LAYER_CONFIG_FILE_NAME), 'r') as file:
        layer_config = json.load(file)
    validate_params(layer_root, layer_config, ['name', 'deployment_package'])
    artifact_name = without_zip_ext(layer_config['deployment_package'])
    artifact_path = Path(bundle_dir, artifact_name)

    _LOG.info(f"Going to assemble lambda layer '{layer_config['name']}'")
    package_name = zip_ext(layer_config['deployment_package'])

    r_hash_name = f'.{artifact_name}_{REQ_HASH_SUFFIX}'
    artifact_cache_path = Path(cache_dir_path, artifact_name)

    prev_req_hash = None
    req_hash_path = Path(cache_dir_path, r_hash_name)
    if os.path.exists(req_hash_path):
        with open(req_hash_path, 'r') as f:
            prev_req_hash = f.read().strip()

    # install requirements.txt content
    requirements_path = Path(layer_root, REQ_FILE_NAME)
    if os.path.exists(requirements_path):
        current_req_hash = compute_file_hash(requirements_path)
        if (current_req_hash != EMPTY_FILE_HASH and
                prev_req_hash != current_req_hash):
            _LOG.debug(f'Artifacts cache path: {artifact_cache_path}')
            os.makedirs(artifact_cache_path, exist_ok=True)

            install_requirements_to(requirements_path, to=artifact_cache_path,
                                    config=layer_config,
                                    errors_allowed=errors_allowed)

            _LOG.debug('Zipping 3-rd party dependencies')
            zip_dir(str(artifact_cache_path),
                    str(Path(cache_dir_path, package_name)))

            remove_dir(artifact_cache_path)
            _LOG.debug(f'"{artifact_cache_path}" was removed successfully')

            with open(req_hash_path, 'w') as f:
                f.write(current_req_hash)
                _LOG.debug(f'Updated requirements hash to {current_req_hash}')
        else:
            _LOG.info(
                f"Skipping installation from the '{requirements_path}' "
                f"because no changes detected from the previous run")

    # install local requirements
    local_requirements_path = Path(layer_root, LOCAL_REQ_FILE_NAME)
    if os.path.exists(local_requirements_path):
        tmp_artifact_path = Path(bundle_dir, artifact_name, TMP_DIR)

        _LOG.info(f'Artifacts path: {tmp_artifact_path}')
        os.makedirs(tmp_artifact_path, exist_ok=True)

        _LOG.info('Going to install local dependencies')
        _install_local_req(tmp_artifact_path, local_requirements_path,
                           project_path)
        _LOG.info('Local dependencies were installed successfully')

        # making zip archive
        _LOG.info(
            f'Packaging artifacts by {tmp_artifact_path} to {package_name}')
        zip_dir(str(tmp_artifact_path),
                str(Path(artifact_path, package_name)))

    if (Path(cache_dir_path, package_name).exists() or
            Path(artifact_path, package_name).exists()):
        _LOG.info(f"Merging lambda layer code with 3-rd party dependencies")
        merge_zip_files(str(Path(artifact_path, package_name)),
                        str(Path(cache_dir_path, package_name)),
                        str(Path(bundle_dir, package_name)),
                        output_subfolder=PYTHON_LAMBDA_LAYER_PATH)
    else:
        raise ArtifactAssemblingError(
            f"Layer package cannot be empty. "
            f"Please check the layer '{layer_config['name']}' configuration.")

    _LOG.info(f'Package \'{package_name}\' was successfully created')

    # remove unused folder
    remove_dir(artifact_path)
    _LOG.info(f'"{artifact_path}" was removed successfully')


@unpack_kwargs
def _build_python_artifact(root, config_file, target_folder, project_path,
                           errors_allowed):
    _LOG.info(f'Building artifact in {target_folder}')

    cache_dir_path = resolve_bundles_cache_directory()
    os.makedirs(cache_dir_path, exist_ok=True)

    # create folder to store artifacts
    with open(config_file, 'r') as file:
        lambda_config = json.load(file)
    validate_params(root, lambda_config, ['lambda_path', 'name', 'version'])

    lambda_name = lambda_config['name']
    artifact_name = f'{lambda_name}-{lambda_config["version"]}'
    artifact_path = Path(target_folder, artifact_name)
    tmp_artifact_path = Path(target_folder, artifact_name, TMP_DIR)

    _LOG.info(f"Going to assemble lambda '{lambda_name}'")
    package_name = build_py_package_name(lambda_name, lambda_config["version"])

    r_hash_name = f'.{artifact_name}_{REQ_HASH_SUFFIX}'
    artifact_cache_path = Path(cache_dir_path, artifact_name)

    _LOG.info(f'Artifacts path: {artifact_path}')
    os.makedirs(tmp_artifact_path, exist_ok=True)

    prev_req_hash = None
    req_hash_path = Path(cache_dir_path, r_hash_name)
    if os.path.exists(req_hash_path):
        with open(req_hash_path, 'r') as f:
            prev_req_hash = f.read().strip()

    # install requirements.txt content
    requirements_path = Path(root, REQ_FILE_NAME)
    if os.path.exists(requirements_path):
        current_req_hash = compute_file_hash(requirements_path)
        if (current_req_hash != EMPTY_FILE_HASH and
                prev_req_hash != current_req_hash):
            _LOG.debug(f'Artifacts cache path: {artifact_cache_path}')
            os.makedirs(artifact_cache_path, exist_ok=True)
            
            install_requirements_to(requirements_path, to=artifact_cache_path,
                                    config=lambda_config,
                                    errors_allowed=errors_allowed)
            _LOG.debug(
                f'Zipping 3-rd party dependencies in {artifact_cache_path}')
            zip_dir(str(artifact_cache_path), 
                    str(Path(cache_dir_path, package_name)))

            remove_dir(artifact_cache_path)
            _LOG.debug(f'"{artifact_cache_path}" was removed successfully')
            
            with open(req_hash_path, 'w') as f:
                f.write(current_req_hash)
                _LOG.debug(f'Updated requirements hash to {current_req_hash}')
        else:
            _LOG.info(
                f"Skipping installation from the '{requirements_path}' "
                f"because no changes detected from the previous run")

    # install local requirements
    local_requirements_path = Path(root, LOCAL_REQ_FILE_NAME)
    if os.path.exists(local_requirements_path):
        _LOG.info('Going to install local dependencies')
        _install_local_req(tmp_artifact_path, local_requirements_path,
                           project_path)
        _LOG.info('Local dependencies were installed successfully')

    # copy lambda's specific packages
    packages_dir = tmp_artifact_path / 'lambdas' / Path(root).name
    os.makedirs(packages_dir, exist_ok=True)
    for package in filter(
            is_python_package,
            [Path(root, item) for item in os.listdir(root)]):
        _LOG.info(f'Copying package {package} to lambda\'s artifacts packages '
                  f'dir: {packages_dir}')
        shutil.copytree(str(package), str(packages_dir / package.name))
        _LOG.info('Copied successfully')

    # copy lambda's handler to artifacts folder
    _LOG.info(f'Copying lambda\'s handler from {root} to {tmp_artifact_path}')
    _copy_py_files(root, tmp_artifact_path)

    # making zip archive
    _LOG.info(f'Packaging artifacts by {tmp_artifact_path} to {package_name}')
    zip_dir(str(tmp_artifact_path), str(Path(artifact_path, package_name)))

    if Path(cache_dir_path, package_name).exists():
        _LOG.info(f"Merging lambda's '{lambda_name}' code with 3-rd party "
                  f"dependencies")
        merge_zip_files(str(Path(artifact_path, package_name)),
                        str(Path(cache_dir_path, package_name)),
                        str(Path(target_folder, package_name)))
    else:
        _LOG.info('Copying lambda\'s code to target folder')
        shutil.copy2(str(Path(artifact_path, package_name)),
                     str(Path(target_folder, package_name)))

    _LOG.info(f'Package \'{package_name}\' was successfully created')

    # remove unused folder
    remove_dir(artifact_path)
    _LOG.info(f'"{artifact_path}" was removed successfully')


def install_requirements_to(requirements_txt: Union[str, Path],
                            to: Union[str, Path],
                            config: Optional[dict] = None,
                            errors_allowed: bool = False):
    """
    1. If there is NO "platform" parameter in lambda_config.json, then the
    dependency installation will be executed by the default command:
    "pip install -r requirements.txt".

    2. If there is NO "platform" parameter in lambda_config.json and flag
    --errors_allowed is True, dependency installation will be tried by
    executing the default command: "pip install -r requirements.txt" in case
    of failures, installation of dependencies will be performed separately
    for each dependency using the default command:
    "pip install <package1>
    pip install <packageN>".

    3. If there is "platform" parameter in lambda_config.json and flag
    --errors_allowed is False, then the dependency installation will be
    executed by the default command using additional parameters:
    "pip install -r requirements.txt --platform manylinux2014_x86_64
    --only-binary=:all: --implementation=cp --python-version 3.8".

    4. If there is "platform" parameter in lambda_config.json and flag
    --errors_allowed is True, dependency installation will be tried by the
    default command using additional parameters:
    "pip install -r requirements.txt --platform manylinux2014_x86_64
    --only-binary=:all: --implementation=cp --python-version 3.8",
    in case of failures, installation of dependencies will be performed
    separately for each dependency using the default command using additional
    parameters. Dependencies that do not have a specified platform will be
    installed with the --platform=any:
    "pip install <package1> --platform manylinux2014_x86_64
    --only-binary=:all: --implementation=cp --python-version 3.8
    pip install <packageN> --platform manylinux2014_x86_64
    --only-binary=:all: --implementation=cp --python-version 3.8
    pip install <packageN+1>".
    """

    exit_code = None
    config = config or {}
    _LOG.info('Going to install 3-rd party dependencies')
    if platforms := config.get('platforms', []):
        if isinstance(platforms, str):
            platforms = [platforms]
        if not isinstance(platforms, list):
            raise InvalidTypeError(
                'Lambda function parameter \'platforms\' must be type of list')
    supported_platforms = update_platforms(set(platforms))
    python_version = _get_python_version(lambda_config=config)
    if supported_platforms:
        command = build_pip_install_command(  # default installation
            requirement=requirements_txt,
            to=to,
            platforms=supported_platforms,
            python=python_version,
            only_binary=':all:',
            implementation='cp'
        )
        result = subprocess.run(command, capture_output=True, text=True)
        _LOG.info(f'\n{result.stdout}\n{result.stderr}')
        if result.returncode != 0 and errors_allowed:
            # tries to install packages compatible with specific platforms
            # independently
            _LOG.info(
                f'Going to install 3-rd party dependencies for platforms: '
                f'{",".join(supported_platforms)}')
            failed_requirements = install_requirements_independently(
                requirements=requirements_txt,
                to=to,
                supported_platforms=supported_platforms,
                python_version=python_version
            )
            failed_requirements = install_requirements_independently(
                requirements=failed_requirements,
                to=to
            )

            _LOG.info(f'\n{result.stdout}\n{result.stderr}')
            if failed_requirements:
                message = (f'An error occurred while installing '
                           f'requirements: "{failed_requirements}" for '
                           f'package "{to}"')
                _LOG.error(message)
                raise ArtifactAssemblingError(message)
        elif result.returncode != 0:
            exit_code = result.returncode
    else:
        _LOG.info('Installing all the requirements with defaults')
        command = build_pip_install_command(
            requirement=requirements_txt,
            to=to
        )
        result = subprocess.run(command, capture_output=True, text=True)
        exit_code = result.returncode

        if result.returncode != 0 and errors_allowed:
            # tries to install packages independently
            _LOG.info(
                'Installing the requirements with defaults independently')
            failed_requirements = install_requirements_independently(
                requirements=requirements_txt,
                to=to
            )

            _LOG.info(f'\n{result.stdout}\n{result.stderr}')
            if failed_requirements:
                message = (f'An error occurred while installing '
                           f'requirements: "{failed_requirements}" for '
                           f'package "{to}"')
                _LOG.error(message)
                raise ArtifactAssemblingError(message)

    if exit_code:
        message = (f'An error: \n"{result.stdout}\n{result.stderr}"\noccurred '
                   f'while installing requirements: "{str(requirements_txt)}" '
                   f'for package "{to}"\nUse --errors-allowed flag to ignore '
                   f'failures in dependencies installation.')
        _LOG.error(message)
        raise ArtifactAssemblingError(message)
    if exit_code == 0:
        _LOG.info(f'\n{result.stdout}\n{result.stderr}')
    _LOG.info('3-rd party dependencies were installed successfully')


def build_pip_install_command(
        requirement: Optional[Union[str, Path]],
        to: Optional[Union[str, Path]] = None,
        implementation: Optional[str] = None,
        python: Optional[str] = None,
        only_binary: Optional[str] = None,
        platforms: Optional[Set[str]] = None,
        additional_args: Optional[List[str]] = None) -> List[str]:
    """
    :param requirement: path to requirements.txt or just one requirement.
    If the path is not file or does not exist it will be treated as one
    requirement.
    :param to: Optional[str] the path where to install requirements
    :param implementation: Optional[str], can be `cp`
    :param python: Optional[str], can be `3.8`, `3.9`
    :param only_binary: Optional[str], can be `:all:` or `:none:`
    :param platforms: Optional[Set], can be {'manylinux2014_x86_64'}
    :param additional_args: Optional[List[str]] list or some additional args
    :return: List[str]
    """
    command = [
        sys.executable, '-m', 'pip', 'install'
    ]
    r_path = Path(requirement)
    if r_path.exists() and r_path.is_file():
        command.extend(['-r', str(r_path)])
    else:  # not a path to requirements.txt but one requirement
        command.append(requirement)

    if to:
        command.extend(['-t', str(to)])
    if implementation:
        command.extend(['--implementation', 'cp'])
    if python:
        command.extend(['--python-version', python])
    if only_binary:
        command.append(f'--only-binary={only_binary}')
    if platforms:
        command.extend(chain.from_iterable(
            ('--platform', p) for p in platforms
        ))
    command.extend(additional_args or [])
    return command


def update_platforms(platforms: Set[str]) -> Set[str]:
    """
    If platforms are not empty, just return them without changing
    (user's choice). But in case the set is empty and the current
    processor is ARM (mac m1) or OS is Windows, we add
    manylinux2014_x86_64 to the list of platforms because by default
    lambdas with x86_64.
    This code is experimental and can be adjusted
    """
    if platforms:
        return platforms
    _arm = platform.processor() == 'arm'
    _win = platform.system() == 'Windows'
    if _arm or _win:
        platforms.add(MANY_LINUX_2014_PLATFORM)
    return platforms


def install_requirements_independently(requirements: Union[str, Path, List[str]],
                                       to: Union[str, Path],
                                       python_version: str = None,
                                       supported_platforms: Set[str] = None) \
        -> List[str]:
    if type(requirements) != list:
        fp = open(requirements, 'r')
        it = (
            line.split(' #')[0].strip() for line in
            filter(lambda line: not line.strip().startswith('#')
                                and line not in EMPTY_LINE_CHARS, fp)
        )
    else:
        it = requirements
    failed_requirements = []
    implementation = 'cp' if python_version or supported_platforms else None
    only_binary = ':all:' if python_version or supported_platforms else None

    for requirement in it:
        command = build_pip_install_command(
            requirement=requirement,
            to=to,
            implementation=implementation,
            python=python_version,
            only_binary=only_binary,
            platforms=supported_platforms
        )
        result = subprocess.run(command, capture_output=True, text=True)
        if result.returncode != 0:
            message = (f'An error occurred while installing requirements from '
                       f'"{str(requirements)}" for platforms: '
                       f'{",".join(supported_platforms) if supported_platforms else "any"}: '
                       f'for package "{requirement}"'
                       f'\nDetails: \n"{result.stdout}\n{result.stderr}"\n')
            USER_LOG.error(f"\033[93m{message}\033[0m")
            failed_requirements.append(requirement)
        else:
            _LOG.info(f'\n{result.stdout}\n{result.stderr}')
    if type(requirements) != list:
        fp.close()
    return failed_requirements


def _install_local_req(artifact_path, local_req_path, project_path):
    from syndicate.core import CONFIG
    with open(local_req_path) as f:
        local_req_list = f.readlines()
    local_req_list = [path_resolver(r.strip()) for r in local_req_list]
    _LOG.info(f'Installing local dependencies: {local_req_list}')
    # copy folders
    for lrp in local_req_list:
        _LOG.info(f'Processing local dependency: {lrp}')
        shutil.copytree(str(Path(CONFIG.project_path, project_path, lrp)),
                        str(Path(artifact_path, lrp)))
        _LOG.debug('Dependency was copied successfully')

        folders = [r for r in lrp.split(DEFAULT_SEP) if r]
        # process folder from root python project
        folders.insert(0, '')
        i = 0
        temp_path = ''
        while i < len(folders):
            temp_path += DEFAULT_SEP + folders[i]
            src_path = Path(CONFIG.project_path, project_path, temp_path)
            dst_path = Path(artifact_path, temp_path)
            _copy_py_files(str(src_path), str(dst_path))
            i += 1
        _LOG.debug('Python files from packages were copied successfully')


def _get_python_version(lambda_config: dict) -> Optional[str]:
    """
    Lambda config or layer config.
     "runtime": "python3.7" => "3.7".
    If "runtime" contains a list with runtimes. The lowest version is returned
    """
    runtimes: Union[None, List, str] = lambda_config.get('runtime')
    if not runtimes:
        return
    if isinstance(runtimes, str):
        runtimes = [runtimes]
    return sorted(
        ''.join(ch for ch in runtime if ch.isdigit() or ch == '.')
        for runtime in runtimes
    )[0]


def _copy_py_files(search_path, destination_path):
    files = glob.iglob(build_path(search_path, _PY_EXT))
    for py_file in files:
        if os.path.isfile(py_file):
            shutil.copy2(py_file, destination_path)


def is_python_package(path: Union[Path, str]) -> bool:
    """A file is considered to be a package if it's a directory containing
    __init__.py"""
    return os.path.isdir(path) and os.path.exists(Path(path, '__init__.py'))



"""
    Copyright 2024 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import io
import json
import os
import posixpath
from functools import cached_property
from pathlib import PurePath
from shutil import rmtree
from typing import Optional
from zipfile import ZipFile

from syndicate.commons import deep_get
from syndicate.exceptions import ResourceNotFoundError, ParameterError
from syndicate.commons.log_helper import get_logger, get_user_logger
from syndicate.core.constants import ARTIFACTS_FOLDER, S3_PATH_NAME, \
    SWAGGER_UI_SPEC_NAME_TEMPLATE, API_GATEWAY_TYPE, RESOURCE_TYPE_PARAM, \
    RESOURCE_NAME_PARAM, PARAMETER_NAME_PARAM, PARAMETER_TYPE_PARAM
from syndicate.core.helper import build_path, unpack_kwargs
from syndicate.core.resources.base_resource import BaseResource
from syndicate.core.resources.helper import build_description_obj

INDEX_FILE_NAME = 'index.html'
X_SYNDICATE_SERVER_PARAM = 'x-syndicate-server'
# example:
# "x-syndicate-server": {
#     "resource_name": "sdct-at-api-gw",
#     "resource_type": "api_gateway",
#     "parameter_name": "stage_name"
#     "parameter_name": "api"
#   }

S3_LINK_TEMPLATE = 'http://{target_bucket}.s3-website.{region}.amazonaws.com'
API_LINK_TEMPLATE = 'https://{api_id}.execute-api.{region}.amazonaws.com/{stage_name}'

_LOG = get_logger(__name__)
USER_LOG = get_user_logger()


class SwaggerUIResource(BaseResource):

    def __init__(self, s3_conn, deploy_target_bucket,
                 deploy_target_bucket_key_compound, region,
                 account_id, extended_prefix_mode, prefix, suffix) -> None:
        from syndicate.core import CONF_PATH, RESOURCES_PROVIDER
        self.s3_conn = s3_conn
        self.deploy_target_bucket = deploy_target_bucket
        self.deploy_target_bucket_key_compound = \
            deploy_target_bucket_key_compound
        self.region = region
        self.account_id = account_id
        self.extended_prefix_mode = extended_prefix_mode
        self.prefix = prefix
        self.suffix = suffix
        self.conf_path = CONF_PATH
        self.resources_provider = RESOURCES_PROVIDER

    @cached_property
    def server_api_mapping(self):
        return {
            API_GATEWAY_TYPE: self._get_api_gateway_link
        }

    def create_update_swagger_ui(self, args):
        return self.create_pool(self._create_update_swagger_ui_from_meta, args)

    def remove_swagger_ui(self, args):
        return self.create_pool(self._remove_swagger_ui, args)

    @unpack_kwargs
    def _create_update_swagger_ui_from_meta(self, name, meta, context=None):
        artifact_dir = PurePath(self.conf_path, ARTIFACTS_FOLDER,
                                name).as_posix()
        target_bucket = meta.get('target_bucket')
        if not target_bucket:
            raise ParameterError(
                f'Target bucket for Swagger UI \'{name}\' is absent in '
                f'resource meta'
            )
        if not self.s3_conn.is_bucket_exists(target_bucket):
            raise ResourceNotFoundError(
                f'Target bucket \'{target_bucket}\' for Swagger UI \'{name}\' '
                f'doesn\'t exists'
            )
        artifact_path = meta.get(S3_PATH_NAME)
        if not artifact_path:
            raise ParameterError(f'Can\'t resolve Swagger UI artifact path')
        artifact_src_path = posixpath.join(
            self.deploy_target_bucket_key_compound, artifact_path)

        _LOG.info(f'Downloading an artifact for Swagger UI \'{name}\'')
        with io.BytesIO() as artifact:
            self.s3_conn.download_to_file(
                bucket_name=self.deploy_target_bucket,
                key=artifact_src_path,
                file=artifact)
            extract_to = build_path(artifact_dir, name)
            with ZipFile(artifact, 'r') as zf:
                zf.extractall(extract_to)

        _LOG.info(f'Uploading files for Swagger UI \'{name}\' to target '
                  f'bucket \'{target_bucket}\'')
        for file in os.listdir(extract_to):
            extra_args = None
            if file == INDEX_FILE_NAME:
                extra_args = {
                    'ContentType': 'text/html',
                    'ContentDisposition': f'inline;filename={INDEX_FILE_NAME}'
                }
            elif file.endswith('.json'):
                filepath = PurePath(extract_to, file)
                self.resolve_api_url(filepath, meta)

            self.s3_conn.upload_single_file(path=PurePath(extract_to,
                                                          file).as_posix(),
                                            key=file,
                                            bucket=target_bucket,
                                            extra_args=extra_args)

        _LOG.info(f'Removing temporary directory \'{artifact_dir}\'')
        rmtree(artifact_dir)

        return self.describe_swagger_ui(name, meta)

    def describe_swagger_ui(self, name, meta):
        target_bucket = meta.get('target_bucket')
        arn = (f'arn:aws-syndicate:{self.region}:{self.account_id}:'
               f'{name}')
        spec_file_name = SWAGGER_UI_SPEC_NAME_TEMPLATE.format(
            name=name)
        website_hosting = self.s3_conn.get_bucket_website(target_bucket)
        bucket_description = {
            'arn': f'arn:aws:s3:::{target_bucket}',
            'bucket_acl': self.s3_conn.get_bucket_acl(target_bucket),
            'location': self.s3_conn.get_bucket_location(target_bucket),
            'policy': self.s3_conn.get_bucket_policy(target_bucket)
        }
        if not bucket_description['location']:
            return {}
        response = {
            'host_description': bucket_description
        }
        if website_hosting:
            hosting_config = {
                "enabled": True,
                "index_document": deep_get(website_hosting,
                                           ['IndexDocument', 'Suffix'])
            }
            if self.s3_conn.is_file_exists(target_bucket, spec_file_name):
                hosting_config['api_spec_document'] = spec_file_name
            hosting_config['endpoint'] = (
                S3_LINK_TEMPLATE.format(target_bucket=target_bucket,
                                        region=self.region)
            )
            response['website_hosting'] = hosting_config
        return {
            arn: build_description_obj(response, name, meta)
        }

    @unpack_kwargs
    def _remove_swagger_ui(self, arn, config) -> dict:
        resource_name = arn.split(':')[-1]
        target_bucket = deep_get(config, ['resource_meta', 'target_bucket'])

        pure_name = resource_name
        if self.extended_prefix_mode:
            if self.prefix:
                pure_name = pure_name[len(self.prefix):]
            if self.suffix:
                pure_name = pure_name[:-len(self.suffix)]

        if not self.s3_conn.is_bucket_exists(target_bucket):
            USER_LOG.info(f'Target bucket with name \'{target_bucket}\' not '
                          f'found')
            return {arn: config}
        else:
            self.s3_conn.remove_object(bucket_name=target_bucket,
                                       file_name=INDEX_FILE_NAME)
            self.s3_conn.remove_object(
                bucket_name=target_bucket,
                file_name=SWAGGER_UI_SPEC_NAME_TEMPLATE.format(
                    name=pure_name))
            USER_LOG.info(f'Swagger UI \'{resource_name}\' removed')
            return {arn: config}

    @staticmethod
    def validate_server_params(server_params: dict) -> bool:
        required = [RESOURCE_TYPE_PARAM,
                    RESOURCE_NAME_PARAM,
                    PARAMETER_NAME_PARAM,
                    PARAMETER_TYPE_PARAM]

        if not set(server_params).issubset(set(required)):
            _LOG.error(f'Incorrect values in `{X_SYNDICATE_SERVER_PARAM}` '
                       f'parameter. Must be a subset of these: {required}')
            USER_LOG.warn('Cannot retrieve active URL for swagger UI')
            return False
        return True

    def _get_api_gateway_link(self, server_params: dict, meta: dict) -> \
            Optional[str]:
        resource_name = server_params.get(RESOURCE_NAME_PARAM)
        if self.extended_prefix_mode:
            if self.prefix:
                resource_name = self.prefix + resource_name
            if self.suffix:
                resource_name = resource_name + self.suffix
        stage_name = server_params.get(PARAMETER_NAME_PARAM)

        response = self.resources_provider.api_gw().\
            describe_api_resources(name=resource_name, meta=meta)
        if not response:
            _LOG.warning(
                f'Cannot find `{resource_name}` API Gateway. Cannot retrieve '
                f'active URL for swagger UI')
            return

        first_value = next(iter(response.values()))
        api_url = API_LINK_TEMPLATE.format(
            api_id=first_value['description']['id'],
            region=self.region,
            stage_name=stage_name
        )
        return api_url

    def resolve_api_url(self, filepath: str | PurePath, meta: dict) -> None:
        try:
            with open(filepath, 'r') as f:
                file_content = json.load(f)

            server_params = file_content.get(X_SYNDICATE_SERVER_PARAM)
            _LOG.debug(f'x-syndicate-server parameters: {server_params}')
            if not server_params or \
                    not self.validate_server_params(server_params):
                return

            resource_type = server_params.get(RESOURCE_TYPE_PARAM)
            api_func = self.server_api_mapping.get(resource_type)

            if not api_func:
                _LOG.warning(
                    f'The resource type \'{resource_type}\' is not supported '
                    f'for \'{X_SYNDICATE_SERVER_PARAM}\' configuration.')
                return

            api_url = api_func(server_params, meta)
            if not api_url:
                return

            file_content['servers'][0]['url'] = api_url
            with open(filepath, 'w') as f:
                json.dump(file_content, f)
        except (FileNotFoundError, KeyError, AttributeError,
                json.JSONDecodeError) as e:
            _LOG.error(f'Error processing file {filepath}: {str(e)}')
